#!/usr/bin/env python
# python script to generate an overview of the states based on the input lex file.
#
# Copyright (C) 1997-2019 by Dimitri van Heesch.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation under the terms of the GNU General Public License is hereby
# granted. No representations are made about the suitability of this software
# for any purpose. It is provided "as is" without express or implied warranty.
# See the GNU General Public License for more details.
#
# Documents produced by Doxygen are derivative works derived from the
# input used in their production; they are not affected by this license.
#
import sys
import os
import re


def main():
    if len(sys.argv)!=2:
        sys.exit('Usage: %s <lex_file>' % sys.argv[0])

    lex_file = sys.argv[1]
    if (os.path.exists(lex_file)):
        #write preamble
        print("static const char *stateToString(int state)")
        print("{")
        print("  switch(state)")
        print("  {")
        print("    case INITIAL: return \"INITIAL\";")

        with open(lex_file) as f:
            for line in f:
                if re.search(r'^%x', line) or  re.search(r'^%s', line):
                    state = line.split()[1]
                    print("    case %s: return \"%s\";" % (state,state))
                elif re.search(r'^%%', line):
                    break
                else:
                    pass
            f.close()
        #write post
        print("  }")
        print("  return \"Unknown\";")
        print("}")

if __name__ == '__main__':
    main()



# An example showing how to use the python doxmlparser module to extract some metrics from
# the XML output generated by doxygen for a project.

import sys

import doxmlparser

from doxmlparser.compound import DoxCompoundKind, DoxMemberKind, DoxSectionKind, MixedContainer

class Metrics:
    def __init__(self):
        self.numClasses = 0
        self.numDocClasses = 0
        self.numStructs = 0
        self.numUnions = 0
        self.numInterfaces = 0
        self.numExceptions = 0
        self.numNamespaces = 0
        self.numFiles = 0
        self.numDocFiles = 0
        self.numGroups = 0
        self.numPages = 0
        self.numPubMethods = 0
        self.numDocPubMethods = 0
        self.numProMethods = 0
        self.numDocProMethods = 0
        self.numPriMethods = 0
        self.numDocPriMethods = 0
        self.numAttributes = 0
        self.numDocAttributes = 0
        self.numFunctions = 0
        self.numDocFunctions = 0
        self.numVariables = 0
        self.numDocVariables = 0
        self.numParams = 0
    def print(self):
        numMethods    = self.numPubMethods    + self.numProMethods    + self.numPriMethods
        numDocMethods = self.numDocPubMethods + self.numDocProMethods + self.numDocPriMethods
        print("Metrics:")
        print("-----------------------------------")
        if self.numClasses>0:
            print("Classes:     {:=10} ({} documented)".format(self.numClasses,self.numDocClasses))
        if self.numStructs>0:
            print("Structs:     {:=10}".format(self.numStructs))
        if self.numUnions>0:
            print("Unions:      {:=10}".format(self.numUnions))
        if self.numExceptions>0:
            print("Exceptions:  {:=10}".format(self.numExceptions))
        if self.numNamespaces>0:
            print("Namespaces:  {:=10}".format(self.numNamespaces))
        if self.numFiles>0:
            print("Files:       {:=10} ({} documented)".format(self.numFiles,self.numDocFiles))
        if self.numGroups>0:
            print("Groups:      {:=10}".format(self.numGroups))
        if self.numPages>0:
            print("Pages:       {:=10}".format(self.numPages))
        if numMethods>0:
            print("Methods:     {:=10} ({} documented)".format(numMethods,numDocMethods))
        if self.numPubMethods>0:
            print("  Public:    {:=10} ({} documented)".format(self.numPubMethods,self.numDocPubMethods))
        if self.numProMethods>0:
            print("  Protected: {:=10} ({} documented)".format(self.numProMethods,self.numDocProMethods))
        if self.numPriMethods>0:
            print("  Private:   {:=10} ({} documented)".format(self.numPriMethods,self.numDocPriMethods))
        if self.numFunctions>0:
            print("Functions:   {:=10} ({} documented)".format(self.numFunctions,self.numDocFunctions))
        if self.numAttributes>0:
            print("Attributes:  {:=10} ({} documented)".format(self.numAttributes,self.numDocAttributes))
        if self.numVariables>0:
            print("Variables:   {:=10} ({} documented)".format(self.numVariables,self.numDocVariables))
        if self.numParams>0:
            print("Params:      {:=10}".format(self.numParams))
        print("-----------------------------------")
        if self.numClasses>0:
            print("Avg. #methods/compound:  {:=10}".format(float(numMethods)/float(self.numClasses)))
        if numMethods>0:
            print("Avg. #params/method:     {:=10}".format(float(self.numParams)/float(numMethods)))
        print("-----------------------------------")


def description_is_empty(description):
    for content in description.content_:
        if content.getCategory()==MixedContainer.CategoryText:
            if not content.getValue().isspace():
                return False # non space-only text
        elif not content.getCategory()==MixedContainer.CategoryNone:
            return False # some internal object like a paragraph
    return True

def is_documented(definition):
    return not description_is_empty(definition.get_briefdescription()) or \
           not description_is_empty(definition.get_detaileddescription())

def section_is_protected(sectionkind):
    return sectionkind in [DoxSectionKind.PROTECTEDTYPE,
                           DoxSectionKind.PROTECTEDFUNC,
                           DoxSectionKind.PROTECTEDATTRIB,
                           DoxSectionKind.PROTECTEDSLOT,
                           DoxSectionKind.PROTECTEDSTATICFUNC,
                           DoxSectionKind.PROTECTEDSTATICATTRIB]

def section_is_private(sectionkind):
    return sectionkind in [DoxSectionKind.PRIVATETYPE,
                           DoxSectionKind.PRIVATEFUNC,
                           DoxSectionKind.PRIVATEATTRIB,
                           DoxSectionKind.PRIVATESLOT,
                           DoxSectionKind.PRIVATESTATICFUNC,
                           DoxSectionKind.PRIVATESTATICATTRIB]

def section_is_public(sectionkind):
    return not section_is_protected(sectionkind) and not section_is_private(sectionkind)

def linked_text_to_string(linkedtext):
    str=''
    if linkedtext:
        for text_or_ref in linkedtext.content_:
            if text_or_ref.getCategory()==MixedContainer.CategoryText:
                str+=text_or_ref.getValue()
            else:
                str+=text_or_ref.getValue().get_valueOf_()
    return str

def parse_members(compounddef,sectiondef,metrics):
    functionLikeKind = [DoxMemberKind.FUNCTION,
                        DoxMemberKind.PROTOTYPE,
                        DoxMemberKind.SIGNAL,
                        DoxMemberKind.SLOT,
                        DoxMemberKind.DCOP]
    variableLikeKind = [DoxMemberKind.VARIABLE, DoxMemberKind.PROPERTY]
    for memberdef in sectiondef.get_memberdef():
        if compounddef.get_kind() in [DoxCompoundKind.CLASS, DoxCompoundKind.STRUCT, DoxCompoundKind.INTERFACE]:
            if memberdef.get_kind() in functionLikeKind:
                if section_is_public(sectiondef.get_kind()):
                    metrics.numPubMethods+=1
                    if is_documented(memberdef):
                        metrics.numDocPubMethods+=1
                elif section_is_protected(sectiondef.get_kind()):
                    metrics.numProMethods+=1
                    if is_documented(memberdef):
                        metrics.numDocProMethods+=1
                elif section_is_private(sectiondef.get_kind()):
                    metrics.numPriMethods+=1
                    if is_documented(memberdef):
                        metrics.numDocPriMethods+=1
            elif memberdef.get_kind() in variableLikeKind:
                metrics.numAttributes+=1
                if is_documented(memberdef):
                    metrics.numDocAttributes+=1
        elif compounddef.get_kind() in [DoxCompoundKind.FILE, DoxCompoundKind.NAMESPACE]:
            if memberdef.get_kind() in functionLikeKind:
                metrics.numFunctions+=1
                if is_documented(memberdef):
                    metrics.numDocFunctions+=1
            elif memberdef.get_kind() in variableLikeKind:
                metrics.numVariables+=1
                if is_documented(memberdef):
                    metrics.numDocVariables+=1
        #for param in memberdef.get_param():
        #    name = ''
        #    if param.get_defname():
        #        name = param.get_defname()
        #    if param.get_declname():
        #        name = param.get_declname()
        #    print("param '{}':'{}'".format(linked_text_to_string(param.get_type()),name))
        metrics.numParams+=len(memberdef.get_param())
        if memberdef.get_type() and memberdef.get_type()!="void" and linked_text_to_string(memberdef.get_type()):
            metrics.numParams+=1  # count non-void return types as well
            #print("returns '{}'".format(linked_text_to_string(memberdef.get_type())))

def parse_sections(compounddef,metrics):
    for sectiondef in compounddef.get_sectiondef():
        parse_members(compounddef,sectiondef,metrics)

def parse_compound(inDirName,baseName,metrics):
    rootObj = doxmlparser.compound.parse(inDirName+"/"+baseName+".xml",True)
    for compounddef in rootObj.get_compounddef():
        kind = compounddef.get_kind()
        if kind==DoxCompoundKind.CLASS:
            metrics.numClasses+=1
            if is_documented(compounddef):
                metrics.numDocClasses+=1
        elif kind==DoxCompoundKind.STRUCT:
            metrics.numStructs+=1
        elif kind==DoxCompoundKind.UNION:
            metrics.numUnions+=1
        elif kind==DoxCompoundKind.INTERFACE:
            metrics.numInterfaces+=1
        elif kind==DoxCompoundKind.EXCEPTION:
            metrics.numExceptions+=1
        elif kind==DoxCompoundKind.NAMESPACE:
            metrics.numNamespaces+=1
        elif kind==DoxCompoundKind.FILE:
            metrics.numFiles+=1
            if is_documented(compounddef):
                metrics.numDocFiles+=1
        elif kind==DoxCompoundKind.GROUP:
            metrics.numGroups+=1
        elif kind==DoxCompoundKind.PAGE:
            metrics.numPages+=1
        else:
            continue
        parse_sections(compounddef,metrics)

def parse_index(inDirName):
    metrics = Metrics()
    rootObj = doxmlparser.index.parse(inDirName+"/index.xml",True)
    for compound in rootObj.get_compound(): # for each compound defined in the index
        print("Processing {0}...".format(compound.get_name()))
        parse_compound(inDirName,compound.get_refid(),metrics)
    metrics.print()

def usage():
    print("Usage {0} <xml_output_dir>".format(sys.argv[0]))
    sys.exit(1)

def main():
    args = sys.argv[1:]
    if len(args)==1:
        parse_index(args[0])
    else:
        usage()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# python script to generate configoptions.cpp and config.doc from config.xml
#
# Copyright (C) 1997-2015 by Dimitri van Heesch.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation under the terms of the GNU General Public License is hereby
# granted. No representations are made about the suitability of this software
# for any purpose. It is provided "as is" without express or implied warranty.
# See the GNU General Public License for more details.
#
# Documents produced by Doxygen are derivative works derived from the
# input used in their production; they are not affected by this license.
#
import xml.dom.minidom
import sys
import re
import textwrap
from xml.dom import Node
import io
import glob

messages = {}

# wrapper class to write to file/output in UTF-8 format
class OutputWriter:
    def __init__(self,writer) :
        self.writer = io.open(writer.fileno(), 'w', encoding='utf8')

    def write(self, text) :
        if sys.version_info.major == 2:
            self.writer.write(unicode(text))
        else:
            self.writer.write(text)

    def flush(self):
        self.writer.flush()


def transformDocs(doc):
    # join lines, unless it is an empty line
    # remove doxygen layout constructs
        # Note: also look at expert.cpp of doxywizard for doxywizard parts
    doc = doc.strip()
    doc = doc.replace("\n", " ")
    doc = doc.replace("\r", " ")
    doc = doc.replace("\t", " ")
    doc = doc.replace("\\&", "&")
    doc = doc.replace("(\\c ", "(")
    doc = doc.replace("\\c ", " ")
    doc = doc.replace("\\b ", " ")
    doc = doc.replace("\\e ", " ")
    doc = doc.replace("\\$", "$")
    doc = doc.replace("\\#include ", "#include ")
    doc = doc.replace("\\#undef ", "#undef ")
    doc = doc.replace("\\# ", "# ")
    doc = doc.replace("-# ", "\n - ")
    doc = doc.replace(" - ", "\n - ")
    doc = doc.replace("\\sa ", "\nSee also: ")
    doc = doc.replace("\\par ", "\n")
    doc = doc.replace("@note ", "\nNote: ")
    doc = doc.replace("\\note ", "\nNote: ")
    doc = doc.replace("\\verbatim", "\n")
    doc = doc.replace("\\endverbatim", "\n")
    doc = doc.replace("<code>", "")
    doc = doc.replace("</code>", "")
    doc = doc.replace("`", "")
    doc = doc.replace("\\<", "<")
    doc = doc.replace("\\>", ">")
    doc = doc.replace("\\@", "@")
    doc = doc.replace("\\\\", "\\")
    # \ref name "description" -> description
    doc = re.sub('\\\\ref +[^ ]* +"\\\\ref"', '\\\\REF', doc)
    doc = re.sub('\\\\ref +[^ ]* +"([^"]*)"', '\\1', doc)
    # \ref specials
    # \ref <key> -> description
    doc = re.sub('\\\\ref +doxygen_usage', '"Doxygen usage"', doc)
    doc = re.sub('\\\\ref +extsearch', '"External Indexing and Searching"',
                 doc)
    doc = re.sub('\\\\ref +layout', '"Changing the layout of pages"', doc)
    doc = re.sub('\\\\ref +external', '"Linking to external documentation"',
                 doc)
    doc = re.sub('\\\\ref +doxygen_finetune', '"Fine-tuning the output"',
                 doc)
    doc = re.sub('\\\\ref +formulas', '"Including formulas"', doc)
    # fallback for not handled
    doc = re.sub('\\\\ref ', ' ', doc)
    doc = re.sub('\\\\REF', '\\\\ref', doc)
    #<a href="address">description</a> -> description (see: address)
    doc = re.sub('<a +href="([^"]*)" *>([^<]*)</a>', '\\2 (see: \n\\1)', doc)
    # LaTeX name as formula -> LaTeX
    doc = doc.replace("\\f$\\mbox{\\LaTeX}\\f$", "LaTeX")
    # Other formula's (now just 2) so explicitly mentioned.
    doc = doc.replace("\\f$2^{(16+\\mbox{LOOKUP\\_CACHE\\_SIZE})}\\f$",
                      "2^(16+LOOKUP_CACHE_SIZE)")
    doc = doc.replace("\\f$2^{16} = 65536\\f$", "2^16=65536")
    # remove consecutive spaces
    doc = re.sub(" +", " ", doc)
    # a dirty trick to get an extra empty line in Doxyfile documentation.
    # <br> will be removed later on again, we need it here otherwise splitlines
    # will filter the extra line.
    doc = doc.replace("<br>", "\n<br>\n")
    # a dirty trick to go to the next line in Doxyfile documentation.
    # <br/> will be removed later on again, we need it here otherwise splitlines
    # will filter the line break.
    doc = doc.replace("<br/>", "\n<br/>\n")
    #
    doc = doc.splitlines()
    split_doc = []
    for line in doc:
        split_doc += textwrap.wrap(line, 78)
    # replace \ by \\, replace " by \", and '  ' by a newline with end string
    # and start string at next line
    docC = []
    for line in split_doc:
        if (line.strip() != "<br/>"):
            docC.append(line.strip().replace('\\', '\\\\').
                    replace('"', '\\"').replace("<br>", ""))
    return docC


def collectValues(node):
    values = []
    for n in node.childNodes:
        if (n.nodeName == "value"):
            if n.nodeType == Node.ELEMENT_NODE:
                if n.getAttribute('name') != "":
                    if n.getAttribute('show_docu') != "NO":
                        name = "<code>" + n.getAttribute('name') + "</code>"
                        desc = n.getAttribute('desc')
                        if (desc != ""):
                            name += " " + desc
                        values.append(name)
    return values


def addValues(var, node):
    for n in node.childNodes:
        if (n.nodeName == "value"):
            if n.nodeType == Node.ELEMENT_NODE:
                name = n.getAttribute('name')
                print("  %s->addValue(\"%s\");" % (var, name))


def parseHeader(node,objName):
    doc = ""
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            if (n.nodeName == "docs"):
                if (n.getAttribute('doxyfile') != "0"):
                    doc += parseDocs(n)
    docC = transformDocs(doc)
    print("  %s->setHeader(" % (objName))
    rng = len(docC)
    for i in range(rng):
        line = docC[i]
        if i != rng - 1:  # since we go from 0 to rng-1
            print("              \"%s\\n\"" % (line))
        else:
            print("              \"%s\"" % (line))
    print("             );")


def prepCDocs(node):
    type = node.getAttribute('type')
    format = node.getAttribute('format')
    defval = node.getAttribute('defval')
    #adefval = node.getAttribute('altdefval')
    doc = ""
    if (type != 'obsolete'):
        for n in node.childNodes:
            if (n.nodeName == "docs"):
                if (n.getAttribute('doxyfile') != "0"):
                    if n.nodeType == Node.ELEMENT_NODE:
                        doc += parseDocs(n)
        if (type == 'enum'):
            values = collectValues(node)
            doc += "<br/>" + messages['possible']
            rng = len(values)
            for i in range(rng):
                val = values[i]
                if i == rng - 2:
                    doc += "%s%s"%(val,messages['andtxt'])
                elif i == rng - 1:
                    doc += "%s." % (val)
                else:
                    doc += "%s, " % (val)
            if (defval != ""):
                doc += "<br/>" + messages['defvalcode'].format(defval)
        elif (type == 'int'):
            minval = node.getAttribute('minval')
            maxval = node.getAttribute('maxval')
            doc += messages['minmaxdef'].format(minval, maxval, defval)
        elif (type == 'bool'):
            if (node.hasAttribute('altdefval')):
                doc += "<br/>" + messages['defvaltxt'].format(messages['sysdep'])
            else:
                doc += "<br/>" + messages['defvaltxt'].format("YES" if (defval == "1") else "NO")
        elif (type == 'list'):
            if format == 'string':
                values = collectValues(node)
                rng = len(values)
                for i in range(rng):
                    val = values[i]
                    if i == rng - 2:
                        doc += "%s%s"%(val,messages['andtxt'])
                    elif i == rng - 1:
                        doc += "%s." % (val)
                    else:
                        doc += "%s, " % (val)
        elif (type == 'string'):
            if format == 'dir':
                if defval != '':
                    doc += "<br/>" + messages['defdir'].format(defval)
            elif format == 'file':
                abspath = node.getAttribute('abspath')
                if defval != '':
                    if abspath != '1':
                        doc += "<br/>" + messages['deffile'].format(defval)
                    else:
                        doc += "<br/>" + messages['deffileabs'].format(defval)
                else:
                    if abspath == '1':
                        doc += "<br/>" + messages['deffilefull']
            elif format =='image':
                abspath = node.getAttribute('abspath')
                if defval != '':
                    if abspath != '1':
                        doc += "<br/>" + messages['defimg'].format(defval)
                    else:
                        doc += "<br/>" + messages['defimgabs'].format(defval)
                else:
                    if abspath == '1':
                        doc += "<br/>" + messages['defimgfull']
            else: # format == 'string':
                if defval != '':
                    doc += "<br/>" + messages['defvalcode'].format(defval)
        # depends handling
        if (node.hasAttribute('depends')):
            depends = node.getAttribute('depends')
            doc += "<br/>" + messages['depstxt'].format(depends.lower(), depends.upper())

    docC = transformDocs(doc)
    return docC


def parseOption(node):
    # Handling part for Doxyfile
    name = node.getAttribute('id')
    if len(name)>23:
        raise Exception('Option name {0} too long ({1}, where max is 23 characters)'.format(name,len(name)))
    type = node.getAttribute('type')
    format = node.getAttribute('format')
    defval = node.getAttribute('defval')
    adefval = node.getAttribute('altdefval')
    depends = node.getAttribute('depends')
    setting = node.getAttribute('setting')
    orgtype = node.getAttribute('orgtype')
    docC = prepCDocs(node)
    if len(setting) > 0:
        print("#if %s" % (setting))
    print("  //----")
    if type == 'bool':
        if len(adefval) > 0:
            enabled = adefval
        elif defval == '1':
            enabled = "TRUE"
        else:
            enabled = "FALSE"
        print("  cb = cfg->addBool(")
        print("             \"%s\"," % (name))
        rng = len(docC)
        for i in range(rng):
            line = docC[i]
            if i != rng - 1:  # since we go from 0 to rng-1
                print("              \"%s\\n\"" % (line))
            else:
                print("              \"%s\"," % (line))
        print("              %s" % (enabled))
        print("             );")
        if depends != '':
            print("  cb->addDependency(\"%s\");" % (depends))
    elif type == 'string':
        print("  cs = cfg->addString(")
        print("              \"%s\"," % (name))
        rng = len(docC)
        for i in range(rng):
            line = docC[i]
            if i != rng - 1:  # since we go from 0 to rng-1
                print("              \"%s\\n\"" % (line))
            else:
                print("              \"%s\"" % (line))
        print("             );")
        if defval != '':
            print("  cs->setDefaultValue(\"%s\");" % (defval.replace('\\','\\\\')))
        if format == 'file':
            print("  cs->setWidgetType(ConfigString::File);")
        elif format == 'image':
            print("  cs->setWidgetType(ConfigString::Image);")
        elif format == 'dir':
            print("  cs->setWidgetType(ConfigString::Dir);")
        elif format == 'filedir':
            print("  cs->setWidgetType(ConfigString::FileAndDir);")
        if depends != '':
            print("  cs->addDependency(\"%s\");" % (depends))
    elif type == 'enum':
        print("  ce = cfg->addEnum(")
        print("              \"%s\"," % (name))
        rng = len(docC)
        for i in range(rng):
            line = docC[i]
            if i != rng - 1:  # since we go from 0 to rng-1
                print("              \"%s\\n\"" % (line))
            else:
                print("              \"%s\"," % (line))
        print("              \"%s\"" % (defval))
        print("             );")
        addValues("ce", node)
        if depends != '':
            print("  ce->addDependency(\"%s\");" % (depends))
    elif type == 'int':
        minval = node.getAttribute('minval')
        maxval = node.getAttribute('maxval')
        print("  ci = cfg->addInt(")
        print("              \"%s\"," % (name))
        rng = len(docC)
        for i in range(rng):
            line = docC[i]
            if i != rng - 1:  # since we go from 0 to rng-1
                print("              \"%s\\n\"" % (line))
            else:
                print("              \"%s\"," % (line))
        print("              %s,%s,%s" % (minval, maxval, defval))
        print("             );")
        if depends != '':
            print("  ci->addDependency(\"%s\");" % (depends))
    elif type == 'list':
        print("  cl = cfg->addList(")
        print("              \"%s\"," % (name))
        rng = len(docC)
        for i in range(rng):
            line = docC[i]
            try:
                if i != rng - 1:  # since we go from 0 to rng-1
                    print("              \"%s\\n\"" % (line))
                else:
                    print("              \"%s\"" % (line))
            except Exception as inst:
                sys.stdout = sys.stderr
                print("")
                print(inst)
                print("")
        print("             );")
        addValues("cl", node)
        if depends != '':
            print("  cl->addDependency(\"%s\");" % (depends))
        if format == 'file':
            print("  cl->setWidgetType(ConfigList::File);")
        elif format == 'dir':
            print("  cl->setWidgetType(ConfigList::Dir);")
        elif format == 'filedir':
            print("  cl->setWidgetType(ConfigList::FileAndDir);")
    elif type == 'obsolete':
        print("  cfg->addObsolete(\"%s\",ConfigOption::O_%s);" % (name,orgtype.capitalize()))
    if len(setting) > 0:
        print("#else")
        print("  cfg->addDisabled(\"%s\");" % (name))
        print("#endif")


def parseGroups(node):
    name = node.getAttribute('name')
    doc = node.getAttribute('docs')
    setting = node.getAttribute('setting')
    if len(setting) > 0:
        print("#if %s" % (setting))
    print("%s%s" % ("  //-----------------------------------------",
                    "----------------------------------"))
    print("  cfg->addInfo(\"%s\",\"%s\");" % (name, doc))
    print("%s%s" % ("  //-----------------------------------------",
                    "----------------------------------"))
    if len(setting) > 0:
        print("#endif")
    print("")
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            parseOption(n)

def parseGroupMapEnums(node):
    def escape(value):
        return re.sub(r'[^\w]','_',value)
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            type   = n.getAttribute('type')
            name   = n.getAttribute('id')
            defval = n.getAttribute('defval')
            if type=='enum':
                print("\nenum class %s_t" % (name))
                print("{")
                for nv in n.childNodes:
                    if nv.nodeName == "value":
                        value = nv.getAttribute('name')
                        if value:
                            print("  %s," % (escape(value)))
                print("};\n")
                print("inline {0}_t {1}_str2enum(const QCString &s)".format(name,name))
                print("{")
                print("  QCString lc = s.lower();")
                print("  static const std::unordered_map<std::string,{0}_t> map =".format(name))
                print("  {")
                for nv in n.childNodes:
                    if nv.nodeName == "value":
                        value = nv.getAttribute('name')
                        if value:
                            print("    {{ \"{0}\", {1}_t::{2} }},".format(value.lower(),name,escape(value)))
                print("  };")
                print("  auto it = map.find(lc.str());")
                print("  return it!=map.end() ? it->second : {0}_t::{1};".format(name,escape(defval)))
                print("}\n")
                print("inline QCString {0}_enum2str({1}_t v)".format(name,name))
                print("{")
                print("  switch(v)")
                print("  {")
                for nv in n.childNodes:
                    if nv.nodeName == "value":
                        value = nv.getAttribute('name')
                        if value:
                            print("    case {0}_t::{1}: return \"{2}\";".format(name,escape(value),value))
                print("  }")
                print("  return \"{0}\";".format(defval))
                print("}")

def parseGroupMapGetter(node):
    map = { 'bool':'bool', 'string':'const QCString &', 'int':'int', 'list':'const StringVector &' }
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            setting = n.getAttribute('setting')
            if len(setting) > 0:
                print("#if %s" % (setting))
            type = n.getAttribute('type')
            name = n.getAttribute('id')
            if type=='enum':
                print("    %-22s %-30s const                  { return %s(m_%s); }" % (name+'_t',name+'()',name+'_str2enum',name))
                print("    %-22s %-30s const                  { return m_%s; }" % ('const QCString &',name+'_str()',name))
            elif type in map:
                print("    %-22s %-30s const                  { return m_%s; }" % (map[type],name+'()',name))
            if len(setting) > 0:
                print("#endif")

def parseGroupMapSetter(node):
    map = { 'bool':'bool', 'string':'const QCString &', 'int':'int', 'list':'const StringVector &' }
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            setting = n.getAttribute('setting')
            if len(setting) > 0:
                print("#if %s" % (setting))
            type = n.getAttribute('type')
            name = n.getAttribute('id')
            if type=='enum':
                print("    [[maybe_unused]] %-22s update_%-46s { m_%s = %s(v); return v; }" % (name+'_t',name+'('+name+'_t '+' v)',name,name+'_enum2str'))
            elif type in map:
                print("    [[maybe_unused]] %-22s update_%-46s { m_%s = v; return m_%s; }" % (map[type],name+'('+map[type]+' v)',name,name))
            if len(setting) > 0:
                print("#endif")

def parseGroupMapAvailable(node):
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            setting = n.getAttribute('setting')
            type = n.getAttribute('type')
            name = n.getAttribute('id')
            if type=='enum':
                if len(setting) > 0:
                    print("#if %s" % (setting))
                print("    %-22s isAvailable_%-41s { return v.lower() == %s_enum2str(%s_str2enum(v)).lower(); }" % ('bool',name+'(QCString v)',name,name))
                if len(setting) > 0:
                    print("#endif")

def parseGroupMapVar(node):
    map = { 'bool':'bool', 'string':'QCString', 'enum':'QCString', 'int':'int', 'list':'StringVector' }
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            setting = n.getAttribute('setting')
            if len(setting) > 0:
                print("#if %s" % (setting))
            type = n.getAttribute('type')
            name = n.getAttribute('id')
            if type in map:
                print("    %-12s m_%s;" % (map[type],name))
            if len(setting) > 0:
                print("#endif")

def parseGroupInit(node):
    map = { 'bool':'Bool', 'string':'String', 'enum':'Enum', 'int':'Int', 'list':'List' }
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            setting = n.getAttribute('setting')
            if len(setting) > 0:
                print("#if %s" % (setting))
            type = n.getAttribute('type')
            name = n.getAttribute('id')
            if type in map:
                print("  %-25s = ConfigImpl::instance()->get%s(__FILE__,__LINE__,\"%s\");" % ('m_'+name,map[type],name))
            if len(setting) > 0:
                print("#endif")

def getEnum2BoolMapping(node):
    def escape(value):
        return re.sub(r'[^\w]','_',value)
    mapping = []
    for nv in node.childNodes:
        if nv.nodeName == "value":
            name = nv.getAttribute("name")
            bool_rep = nv.getAttribute("bool_representation")
            if name and bool_rep:
                bool_value = "true" if bool_rep and bool_rep.upper() == 'YES' else "false"
                mapping.append( "{{ \"{0}\", {1} }}".format(escape(name),bool_value))
    return mapping

def parseGroupMapInit(node):
    map = { 'bool':'Bool', 'string':'String', 'enum':'String', 'int':'Int', 'list':'List' }
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            setting = n.getAttribute('setting')
            if len(setting) > 0:
                print("#if %s" % (setting))
            type = n.getAttribute('type')
            name = n.getAttribute('id')
            if type in map:
                if type == "enum":
                    mappingStr = "{%s}" % (', '.join(getEnum2BoolMapping(n)))
                    print("    { %-26s Info{ %-13s &ConfigValues::m_%-23s %s}}," % ('\"'+name+'\",','Info::'+map[type]+',',name+",", mappingStr))
                else:
                    print("    { %-26s Info{ %-13s &ConfigValues::m_%-24s}}," % ('\"'+name+'\",','Info::'+map[type]+',',name))
            if len(setting) > 0:
                print("#endif")

def parseGroupCDocs(node):
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            type = n.getAttribute('type')
            name = n.getAttribute('id')
            docC = prepCDocs(n)
            if type != 'obsolete':
                print("  doc->add(")
                print("              \"%s\"," % (name))
                rng = len(docC)
                for i in range(rng):
                    line = docC[i]
                    if i != rng - 1:  # since we go from 0 to rng-1
                        print("              \"%s\\n\"" % (line))
                    else:
                        print("              \"%s\"" % (line))
                print("          );")

def parseOptionDoc(node, first):
    # Handling part for documentation
    name = node.getAttribute('id')
    type = node.getAttribute('type')
    format = node.getAttribute('format')
    defval = node.getAttribute('defval')
    #adefval = node.getAttribute('altdefval')
    depends = node.getAttribute('depends')
    #setting = node.getAttribute('setting')
    doc = ""
    if (type != 'obsolete'):
        for n in node.childNodes:
            if (n.nodeName == "docs"):
                if (n.getAttribute('documentation') != "0"):
                    if n.nodeType == Node.ELEMENT_NODE:
                        doc += parseDocs(n)
        if (first):
            print(" \\anchor cfg_%s" % (name.lower()))
            print("<dl>")
            print("")
            print("<dt>\\c %s <dd>" % (name))
        else:
            print(" \\anchor cfg_%s" % (name.lower()))
            print("<dt>\\c %s <dd>" % (name))
        print(" \\addindex %s" % (name))
        print(doc)
        if (type == 'enum'):
            values = collectValues(node)
            print("")
            print(messages['possible'])
            rng = len(values)
            for i in range(rng):
                val = values[i]
                if i == rng - 2:
                    print("%s%s" % (val,messages['andtxt']))
                elif i == rng - 1:
                    print("%s." % (val))
                else:
                    print("%s, " % (val))
            if (defval != ""):
                print("")
                print("")
                print(messages['defvalcode'].format(defval))
            print("")
        elif (type == 'int'):
            minval = node.getAttribute('minval')
            maxval = node.getAttribute('maxval')
            print("")
            print("")
            print(messages['minmaxdefcode'].format(minval, maxval,defval))
            print("")
        elif (type == 'bool'):
            print("")
            print("")
            if (node.hasAttribute('altdefval')):
                print(messages['defvaltxt'].format(messages['sysdep']))
            else:
                print(messages['defvalcode'].format("YES" if (defval == "1") else "NO"))
            print("")
        elif (type == 'list'):
            if format == 'string':
                values = collectValues(node)
                rng = len(values)
                for i in range(rng):
                    val = values[i]
                    if i == rng - 2:
                        print("%s%s" % (val,messages['andtxt']))
                    elif i == rng - 1:
                        print("%s." % (val))
                    else:
                        print("%s, " % (val))
            print("")
        elif (type == 'string'):
            if format == 'dir':
                if defval != '':
                    print("")
                    print(messages['defdir'].format(defval))
            elif format == 'file':
                abspath = node.getAttribute('abspath')
                if defval != '':
                    print("")
                    if abspath != '1':
                        print(messages['deffile'].format(defval))
                    else:
                        print(messages['deffileabs'].format(defval))
                else:
                    if abspath == '1':
                        print("")
                        print(messages['deffilefull'])
            elif format =='image':
                abspath = node.getAttribute('abspath')
                if defval != '':
                    print("")
                    if abspath != '1':
                        print(messages['defimg'].format(defval))
                    else:
                        print(messages['defimgabs'].format(defval))
                else:
                    if abspath == '1':
                        print("")
                        print(messages['defimgfull'])
            else: # format == 'string':
                if defval != '':
                    print("")
                    print(messages['defvalcode'].format(defval.replace('\\','\\\\')))
            print("")
        # depends handling
        if (node.hasAttribute('depends')):
            depends = node.getAttribute('depends')
            print("")
            print(messages['depstxt'].format(depends.lower(), depends.upper()))
        return False


def parseGroupsDoc(node):
    name = node.getAttribute('name')
    doc = node.getAttribute('docs')
    print("\\section config_%s %s" % (name.lower(), doc))
    # Start of list has been moved to the first option for better
    # anchor placement
    #  print "<dl>"
    #  print ""
    first = True
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            first = parseOptionDoc(n, first)
    if (not first):
        print("</dl>")


def parseGroupsList(node, commandsList):
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            type = n.getAttribute('type')
            if type != 'obsolete':
                commandsList = commandsList + (n.getAttribute('id'),)
    return commandsList


def parseDocs(node):
    doc = ""
    for n in node.childNodes:
        if n.nodeType == Node.TEXT_NODE:
            doc += n.nodeValue.strip()
        elif n.nodeType == Node.CDATA_SECTION_NODE:
            doc += n.nodeValue.rstrip("\r\n ").lstrip("\r\n")
    #doc += "<br>"
    return doc


def parseHeaderDoc(node):
    doc = ""
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            if (n.nodeName == "docs"):
                if (n.getAttribute('documentation') != "0"):
                    doc += parseDocs(n)
    print(doc)


def parseFooterDoc(node):
    doc = ""
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            if (n.nodeName == "docs"):
                if (n.getAttribute('documentation') != "0"):
                    doc += parseDocs(n)
    print(doc)


def parseGenerator(node):
    for n in node.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            if (n.nodeName == "message"):
                name = n.getAttribute('name')
                doc = ""
                for n1 in n.childNodes:
                    if n1.nodeType == Node.TEXT_NODE:
                        doc += n1.nodeValue.strip()
                    elif n1.nodeType == Node.CDATA_SECTION_NODE:
                        doc += n1.nodeValue.rstrip("\r\n").lstrip("\r\n")
                messages[name] = doc

def collectOptions(elem):
    """Collect all option IDs from config.xml."""
    options = set()
    optionsWithElems = {}
    for group in elem.getElementsByTagName('group'):
        for option in group.getElementsByTagName('option'):
            optionId = option.getAttribute('id')
            optionType = option.getAttribute('type')
            if optionId and optionType!='obsolete':
                options.add(optionId)
                optionsWithElems[optionId] = option
    return (options,optionsWithElems)

def syncLocalizedConfig(elem, translationsDir, autoSync=False):
    """Sync localized config_xx.xml files with original config.xml.

    Args:
        elem: The root element of config.xml
        translationsDir: Path to translations directory
        autoSync: If True, automatically sync; if False, only report differences
    """
    import os
    import shutil

    existingOptions, existingOptionsWithElements = collectOptions(elem)
    print("Found %d active options in config.xml" % len(existingOptions))

    srcDir = os.path.dirname(translationsDir)
    if os.path.basename(srcDir) == 'addon':
        srcDir = os.path.dirname(srcDir)
    srcDir = os.path.join(srcDir, 'src')

    for configFile in sorted(glob.glob("translations/config_*.xml")):

        if not os.path.exists(configFile):
            print("Skipping %s: config file not found" % configFile)
            continue

        print("Processing %s..." % configFile)

        try:
            with io.open(configFile, 'r', encoding='utf8') as f:
                content = f.read()
            langDoc = xml.dom.minidom.parseString(content)
        except Exception as e:
            print("  Error parsing %s: %s" % (configFile, e))
            continue

        langOptions = set()
        langOptionsWithElements = {}
        for group in langDoc.getElementsByTagName('group'):
            for option in group.getElementsByTagName('option'):
                optionId = option.getAttribute('id')
                if optionId:
                    langOptions.add(optionId)
                    langOptionsWithElements[optionId] = option

        missingOptions = existingOptions - langOptions
        extraOptions = langOptions - existingOptions

        if missingOptions:
            print("  Missing %d options: %s" % (len(missingOptions), ', '.join(sorted(list(missingOptions))[:5])))
            if len(missingOptions) > 5:
                print("  ... and %d more" % (len(missingOptions) - 5))

        if extraOptions:
            print("  Extra %d options (not in original): %s" % (len(extraOptions), ', '.join(sorted(list(extraOptions))[:5])))
            if len(extraOptions) > 5:
                print("  ... and %d more" % (len(extraOptions) - 5))

        if not missingOptions and not extraOptions:
            print("  OK - all options synchronized")
            continue

        if autoSync and (missingOptions or extraOptions):
            print("  Auto-syncing...")

            rootElement = langDoc.documentElement

            for optionId in extraOptions:
                optionElem = langOptionsWithElements[optionId]
                parentGroup = optionElem.parentNode
                parentGroup.removeChild(optionElem)
                print("    Removed: %s" % optionId)

            for optionId in missingOptions:
                optionElem = existingOptionsWithElements[optionId]
                importedElem = langDoc.importNode(optionElem, True)

                parentGroup = None
                for group in rootElement.getElementsByTagName('group'):
                    parentGroup = group
                    break

                if parentGroup:
                    parentGroup.appendChild(importedElem)
                    print("    Added: %s" % optionId)

            backupFile = configFile + ".bak"
            shutil.copy2(configFile, backupFile)

            outputContent = langDoc.toprettyxml(indent='  ', encoding='utf-8')
            outputStr = outputContent.decode('utf-8') if isinstance(outputContent, bytes) else outputContent

            lines = outputStr.split('\n')
            filteredLines = [line for line in lines if line.strip()]
            outputStr = '\n'.join(filteredLines)

            with io.open(configFile, 'w', encoding='utf8') as f:
                f.write(outputStr)

            print("  Backup saved to: %s" % backupFile)
            print("  File updated: %s" % configFile)

    print("\nSync %s!" % ("complete" if not autoSync else "and update complete"))

def main():
    if len(sys.argv)<3 or (sys.argv[1] not in ['-doc','-cpp','-wiz','-maph','-maps','-sync']):
        sys.exit('Usage: %s -doc|-cpp|-wiz|-maph|-maps|-sync config.xml [translations_dir]' % sys.argv[0])
    try:
        configFile = sys.argv[2]
        if sys.version_info.major == 2:
            fh = open(configFile,'r')
        else:
            fh = open(configFile,'r',encoding='utf8')
        sys.stdout = OutputWriter(sys.stdout)
        doc = xml.dom.minidom.parse(fh)
    except Exception as inst:
        sys.stdout = sys.stderr
        print("")
        print(inst)
        print("")
        sys.exit(1)
    elem = doc.documentElement

    for n in elem.childNodes:
        if n.nodeType == Node.ELEMENT_NODE:
            if (n.nodeName == "generator"):
                parseGenerator(n)
    if len(messages)==0:
        sys.exit('<generator> section missing in %s' % configFile)

    if (sys.argv[1] == "-doc"):
        print("/* WARNING: This file is generated!")
        print(" * Do not edit this file, but edit config.xml instead and run")
        print(" * python configgen.py -doc config.xml to regenerate this file!")
        print(" */")
        # process header
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "header"):
                    parseHeaderDoc(n)
        # generate list with all commands
        commandsList = ()
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    commandsList = parseGroupsList(n, commandsList)
        print("\\secreflist")
        for x in sorted(commandsList):
            print("\\refitem cfg_%s %s" % (x.lower(), x))
        print("\\endsecreflist")
        # process groups and options
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    parseGroupsDoc(n)
        # process footers
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "footer"):
                    parseFooterDoc(n)
    elif (sys.argv[1] == "-maph"):
        print("/* WARNING: This file is generated!")
        print(" * Do not edit this file, but edit config.xml instead and run")
        print(" * python configgen.py -map config.xml to regenerate this file!")
        print(" */")
        print("#ifndef CONFIGVALUES_H")
        print("#define CONFIGVALUES_H")
        print("")
        print("#include <string>")
        print("#include <unordered_map>")
        print("#include \"qcstring.h\"")
        print("#include \"containers.h\"")
        print("#include \"settings.h\"")
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if n.nodeName == "group":
                    parseGroupMapEnums(n)
        print("")
        print("class ConfigValues")
        print("{")
        print("  public:")
        print("    static ConfigValues &instance() { static ConfigValues theInstance; return theInstance; }")
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if n.nodeName == "group":
                    parseGroupMapGetter(n)
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if n.nodeName == "group":
                    parseGroupMapSetter(n)
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if n.nodeName == "group":
                    parseGroupMapAvailable(n)
        print("    void init();")
        print("    StringVector fields() const;")
        print("    struct Info")
        print("    {")
        print("      enum Type { Bool, Int, String, List, Unknown };")
        print("      using Enum2BoolMap = std::unordered_map<std::string,bool>;")
        print("      Info(Type t,bool         ConfigValues::*b) : type(t), value(b) {}")
        print("      Info(Type t,int          ConfigValues::*i) : type(t), value(i) {}")
        print("      Info(Type t,QCString     ConfigValues::*s, const Enum2BoolMap &boolMap = {}) : type(t), value(s), m_boolMap(boolMap) {}")
        print("      Info(Type t,StringVector ConfigValues::*l) : type(t), value(l) {}")
        print("      Type type;")
        print("      union Item")
        print("      {")
        print("        Item(bool         ConfigValues::*v) : b(v) {}")
        print("        Item(int          ConfigValues::*v) : i(v) {}")
        print("        Item(QCString     ConfigValues::*v) : s(v) {}")
        print("        Item(StringVector ConfigValues::*v) : l(v) {}")
        print("        bool         ConfigValues::*b;")
        print("        int          ConfigValues::*i;")
        print("        QCString     ConfigValues::*s;")
        print("        StringVector ConfigValues::*l;")
        print("      } value;")
        print("      bool getBooleanRepresentation() const;")
        print("    private:")
        print("      Enum2BoolMap m_boolMap;")
        print("    };")
        print("    const Info *get(const QCString &tag) const;")
        print("")
        print("  private:")
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    parseGroupMapVar(n)
        print("};")
        print("")
        print("#endif")
    elif (sys.argv[1] == "-maps"):
        print("/* WARNING: This file is generated!")
        print(" * Do not edit this file, but edit config.xml instead and run")
        print(" * python configgen.py -maps config.xml to regenerate this file!")
        print(" */")
        print("#include \"configvalues.h\"")
        print("#include \"configimpl.h\"")
        print("#include <unordered_map>")
        print("")
        print("const ConfigValues::Info *ConfigValues::get(const QCString &tag) const")
        print("{")
        print("  static const std::unordered_map< std::string, Info > configMap =")
        print("  {")
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    parseGroupMapInit(n)
        print("  };")
        print("  auto it = configMap.find(tag.str());")
        print("  return it!=configMap.end() ? &it->second : nullptr;")
        print("}")
        print("")
        print("void ConfigValues::init()")
        print("{")
        print("  static bool first = TRUE;")
        print("  if (!first) return;")
        print("  first = FALSE;")
        print("")
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    parseGroupInit(n)
        print("}")
        print("")
        print("StringVector ConfigValues::fields() const")
        print("{")
        print("  return {")
        first=True
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    for c in n.childNodes:
                        if c.nodeType == Node.ELEMENT_NODE:
                            name = c.getAttribute('id')
                            type = c.getAttribute('type')
                            if type!='obsolete':
                                if not first:
                                    print(",")
                                first=False
                                sys.stdout.write('    "'+name+'"')
        print("")
        print("  };")
        print("}")
        print("")
        print("bool ConfigValues::Info::getBooleanRepresentation() const")
        print("{")
        print("  if (!m_boolMap.empty())")
        print("  {")
        print("    auto it = m_boolMap.find((ConfigValues::instance().*(value.s)).str());")
        print("    if (it!=m_boolMap.end())")
        print("    {")
        print("      return it->second;")
        print("    }")
        print("  }")
        print("  return false;")
        print("}")
        print("")
    elif (sys.argv[1] == "-cpp"):
        print("/* WARNING: This file is generated!")
        print(" * Do not edit this file, but edit config.xml instead and run")
        print(" * python configgen.py -cpp config.xml to regenerate this file!")
        print(" */")
        print("")
        print("#include \"configoptions.h\"")
        print("#include \"configimpl.h\"")
        print("#include \"portable.h\"")
        print("#include \"settings.h\"")
        print("")
        print("void addConfigOptions(ConfigImpl *cfg)")
        print("{")
        print("  ConfigString *cs = nullptr;")
        print("  ConfigEnum   *ce = nullptr;")
        print("  ConfigList   *cl = nullptr;")
        print("  ConfigInt    *ci = nullptr;")
        print("  ConfigBool   *cb = nullptr;")
        print("")
        # process header
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "header"):
                    parseHeader(n,'cfg')
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    parseGroups(n)
        print("}")
    elif (sys.argv[1] == "-wiz"):
        print("/* WARNING: This file is generated!")
        print(" * Do not edit this file, but edit config.xml instead and run")
        print(" * python configgen.py -wiz config.xml to regenerate this file!")
        print(" */")
        print("#include \"configdoc.h\"")
        print("#include \"docintf.h\"")
        print("")
        print("void addConfigDocs(DocIntf *doc)")
        print("{")
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "header"):
                    parseHeader(n,'doc')
        for n in elem.childNodes:
            if n.nodeType == Node.ELEMENT_NODE:
                if (n.nodeName == "group"):
                    parseGroupCDocs(n)
        print("}")
    elif (sys.argv[1] == "-sync"):
        if len(sys.argv) < 3:
            sys.exit('Usage: %s -sync config.xml translations_dir [--auto]' % sys.argv[0])
        translationsDir = sys.argv[3]
        autoSync = '--auto' in sys.argv
        syncLocalizedConfig(elem, translationsDir, autoSync)

if __name__ == '__main__':
    main()



#!/usr/bin/python
# python script to correct the linenumbers due to inclusion of common parts in lex files
#
# Copyright (C) 1997-2022 by Dimitri van Heesch.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation under the terms of the GNU General Public License is hereby
# granted. No representations are made about the suitability of this software
# for any purpose. It is provided "as is" without express or implied warranty.
# See the GNU General Public License for more details.
#
# Documents produced by Doxygen are derivative works derived from the
# input used in their production; they are not affected by this license.
#
import sys
import os
import re

def main():
    if len(sys.argv)!=6:
        sys.exit('Usage: {0} <input_cpp_file> <output_cpp_file> <correction_file> <original_lex_file> <generated_lex_file>'.format(sys.argv[0]))

    inp_cpp_file, out_cpp_file, corr_cpp_file, org_lex, gen_lex = sys.argv[1:]

    quoted_gen_lex      = '"' + gen_lex      + '"'

    corr_list = []
    if (os.path.exists(corr_cpp_file)):
        # read correction file as list of tuples
        with open(corr_cpp_file,"r") as corr_file:
            corr_list = [tuple(map(int, line.split(' '))) for line in corr_file]

    if (os.path.exists(inp_cpp_file)):
        with open(out_cpp_file,"w") as out_file:
            rule_correction = False
            with open(inp_cpp_file) as in_file:
                for line in in_file:
                    # helper function to find the correction needed for the given line number
                    def get_line_correction(line_no):
                        corr_cnt = 0
                        for elem in corr_list:
                            if elem[0] <= line_no:
                                corr_cnt = elem[1]
                        return corr_cnt

                    if re.search(r'^#line ', line): # statement added by C-preprocessing
                        inc_line, inc_file = line.split()[1:3]
                        if inc_file == quoted_gen_lex:
                            line_cnt = int(inc_line)
                            line_cnt -= get_line_correction(line_cnt) # adjust line number
                            out_file.write("#line {0} \"{1}\"\n".format(line_cnt,org_lex))
                        else:
                            out_file.write(line)
                    elif re.search(r'^  /\* #line ', line): # statement added by lex-preprocessing
                        inc_line, inc_file = line.split()[2:4] # first token is `/*` part
                        out_file.write("#line {0} \"{1}\"\n".format(inc_line,inc_file))
                    elif re.search(r'static .* yy_rule_linenum', line): # linenum table generated in debug mode (flex -d)
                        out_file.write(line)
                        rule_correction = True
                        # read next lines till } and correct numbers
                        # Assumption structure is like:
                        #   static const flex_int16_t yy_rule_linenum[26] =
                        #      {   0,
                        #        981,  985,  989,  995, 1093, 1151, 1153, 1154, 1179, 1182,
                        #       1195, 1198, 1201, 1207, 1211, 1214, 1217, 1223, 1226, 1228,
                        #       1230, 1235, 1236, 1237, 1238
                        #      } ;
                    elif rule_correction:
                        if re.search(r'{', line): # start of yy_rule_linenum array
                            out_file.write(line)
                        elif re.search(r'}', line): # end of yy_rule_linenum array
                            out_file.write(line)
                            rule_correction = False
                        else: # inside the table
                            out_file.write("    ")
                            out_list = []
                            rule_list = line.replace(',',' ').split()
                            for rule_cnt in rule_list:
                                rule_num = int(rule_cnt)
                                rule_num -= get_line_correction(rule_num)
                                out_list.append("{0:5d}".format(rule_num))
                            out_file.write(','.join(out_list))
                            if re.search(r',$', line): # re-add trailing comma
                                out_file.write(",")
                            out_file.write("\n")
                    else: # normal line -> just copy
                        out_file.write(line)
    else: # input file does not exist
        sys.exit("Input cpp file '{0}' does not exist.".format(sys.argv[1]))
        exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
from xml.etree import cElementTree as ET
import os
import sqlite3
import sys
import getopt

# map XML attributes/elements to SQL rows
# --POC: iterate through the children and attributes of the memberdef element
#        and search it in doxygen_sqlite3.db

g_conn=None
val=[]
def print_unprocessed_attributes(node):
    for key in node.attrib:
        print("WARNING: '%s' has unprocessed attr '%s'" % (node.tag,key))

def extract_attribute(node,attribute,pnl):
    if attribute not in node.attrib:
        return
    pnl.append("%s = ?" % attribute)
    val.append(node.attrib[attribute])
    node.attrib.pop(attribute)

def extract_element(node,chld,pnl):
    # deal with <tag />
    if chld.text is None:
        if len(chld.attrib)==0:
            node.remove(chld)
        return

    a=chld.text.strip()
    if not a == "":
        pnl.append("%s =?" % chld.tag)
        val.append(chld.text.strip())
    else:
        pnl.append("%s IS NULL OR %s = ''" % (chld.tag,chld.tag))
    node.remove(chld)

def process_memberdef(node):
    q=[]
    for chld in node.getchildren():
        if chld.tag == "referencedby":
            continue
        if chld.tag == "references":
            continue
        if chld.tag == "param":
            continue
        if chld.tag == "type":
            continue
        if chld.tag == "location":
            extract_attribute(chld,"line",q)
            extract_attribute(chld,"column",q)
            extract_attribute(chld,"bodystart",q)
            extract_attribute(chld,"bodyend",q)

            q.append("id_bodyfile=(select id from files where name=?)")
            val.append(chld.attrib["bodyfile"])
            chld.attrib.pop("bodyfile")

            q.append("id_file=(select id from files where name=?)")
            val.append(chld.attrib["file"])
            chld.attrib.pop("file")

            print_unprocessed_attributes(chld)
            if len(chld.attrib) == 0:
                node.remove(chld)
        else:
            extract_element(node,chld,q)

    for chld in node.getchildren():
        print("WARNING: '%s' has unprocessed child elem '%s'" % (node.tag,chld.tag))

    extract_attribute(node,"kind",q)
    extract_attribute(node,"prot",q)
    extract_attribute(node,"static",q)
    extract_attribute(node,"mutable",q)
    extract_attribute(node,"const",q)
    extract_attribute(node,"virt",q)
    extract_attribute(node,"explicit",q)
    extract_attribute(node,"inline",q)

    q.append("refid=?")
    val.append(node.attrib['id'])
    node.attrib.pop('id')

    print_unprocessed_attributes(node)

    query="SELECT * FROM memberdef WHERE %s" % " AND ".join(q)
    r=[]
    try:
        r = g_conn.execute(query,val).fetchall()
    except(sqlite3.OperationalError,e):
        print("SQL_ERROR:%s"%e)

    del val[:]
    if not len(r) > 0:
        print("TEST_ERROR: Member not found in SQL DB")


def load_xml(name):
    context = ET.iterparse(name, events=("start", "end"))
    event, root = context.next()
    for event, elem in context:
        if event == "end" and elem.tag == "memberdef":
            process_memberdef(elem)
    print("\n== Unprocessed XML ==")
#    ET.dump(root)


def open_db(dbname):
    global g_conn

    if dbname is None:
        dbname = "doxygen_sqlite3.db"

    if not os.path.isfile(dbname):
        raise BaseException("No such file %s" % dbname )

    g_conn = sqlite3.connect(dbname)
    g_conn.execute('PRAGMA temp_store = MEMORY;')
    g_conn.row_factory = sqlite3.Row

def main(argv):
    try:
        opts, args = getopt.getopt(argv, "hd:x:",["help"])
    except getopt.GetoptError:
        sys.exit(1)

    dbname=None
    xmlfile=None

    for a, o in opts:
        if a in ('-h', '--help'):
            sys.exit(0)
        elif a in ('-d'):
            dbname=o
            continue
        elif a in ('-x'):
            xmlfile=o
            continue
    open_db(dbname)
    load_xml(xmlfile)

if __name__ == '__main__':
    main(sys.argv[1:])



# called as `replace_version.py src dest v1`
# replaces @VERSION@ by v1 in file src and writes the result to dest
import sys

def main():
    if len(sys.argv)<4:
        print('replace_version.py src dest version')
        sys.exit(1)

    inputFile  = sys.argv[1]
    outputFile = sys.argv[2]
    version    = sys.argv[3]

    with open(outputFile,"w") as out_file:
        with open(inputFile,"r") as in_file:
            for line in in_file:
                out_file.write(line.replace("@VERSION@",version))

if __name__ == '__main__':
    main()



import platform

def is_windows():
    return platform.system().lower() == "windows"

if is_windows():
    from winreg import OpenKey, QueryValueEx, HKEY_LOCAL_MACHINE, KEY_READ

if __name__ == '__main__':
    if is_windows():
        root = HKEY_LOCAL_MACHINE
        subkey = R'SYSTEM\CurrentControlSet\Control\Nls\CodePage'

        key = OpenKey(root, subkey, 0, KEY_READ)
        name = 'ACP'

        try:
            codepage, _ = QueryValueEx(key, name)
            print(codepage)
        except WindowsError:
            print('Failed to get code page')



"""Script to generate reports on translator classes from Doxygen sources.

  The main purpose of the script is to extract the information from sources
  related to internationalization (the translator classes). It uses the
  information to generate documentation (language.dox,
  translator_report.txt) from templates (language.tpl, maintainers.txt).

  Simply run the script without parameters to get the reports and
  documentation for all supported languages. If you want to generate the
  translator report only for some languages, pass their codes as arguments
  to the script. In that case, the language.dox will not be generated.
  Example:

    python translator.py en nl cz

  Originally, the script was written in Perl and was known as translator.pl.
  The last Perl version was dated 2002/05/21 (plus some later corrections)

                                         Petr Prikryl (prikryl at atlas dot cz)

  History:
  --------
  2002/05/21 - This was the last Perl version.
  2003/05/16 - List of language marks can be passed as arguments.
  2004/01/24 - Total reimplementation started: classes TrManager, and Transl.
  2004/02/05 - First version that produces translator report. No language.dox yet.
  2004/02/10 - First fully functional version that generates both the translator
               report and the documentation. It is a bit slower than the
               Perl version, but is much less tricky and much more flexible.
               It also solves some problems that were not solved by the Perl
               version. The translator report content should be more useful
               for developers.
  2004/02/11 - Some tuning-up to provide more useful information.
  2004/04/16 - Added new tokens to the tokenizer (to remove some warnings).
  2004/05/25 - Added from __future__ import generators not to force Python 2.3.
  2004/06/03 - Removed dependency on textwrap module.
  2004/07/07 - Fixed the bug in the fill() function.
  2004/07/21 - Better e-mail mangling for HTML part of language.dox.
             - Plural not used for reporting a single missing method.
             - Removal of not used translator adapters is suggested only
               when the report is not restricted to selected languages
               explicitly via script arguments.
  2004/07/26 - Better reporting of not-needed adapters.
  2004/10/04 - Reporting of not called translator methods added.
  2004/10/05 - Modified to check only doxygen/src sources for the previous report.
  2005/02/28 - Slight modification to generate "mailto.txt" auxiliary file.
  2005/08/15 - Doxygen's root directory determined primarily from DOXYGEN
               environment variable. When not found, then relatively to the script.
  2007/03/20 - The "translate me!" searched in comments and reported if found.
  2008/06/09 - Warning when the MAX_DOT_GRAPH_HEIGHT is still part of trLegendDocs().
  2009/05/09 - Changed HTML output to fit it with XHTML DTD
  2009/09/02 - Added percentage info to the report (implemented / to be implemented).
  2010/02/09 - Added checking/suggestion 'Reimplementation using UTF-8 suggested.
  2010/03/03 - Added [unreachable] prefix used in maintainers.txt.
  2010/05/28 - BOM skipped; minor code cleaning.
  2010/05/31 - e-mail mangled already in maintainers.txt
  2010/08/20 - maintainers.txt to UTF-8, related processing of unicode strings
             - [any mark] introduced instead of [unreachable] only
             - marks highlighted in HTML
  2010/08/30 - Highlighting in what will be the table in langhowto.html modified.
  2010/09/27 - The underscore in \\latexonly part of the generated language.dox
               was prefixed by backslash (was LaTeX related error).
  2013/02/19 - Better diagnostics when translator_xx.h is too crippled.
  2013/06/25 - TranslatorDecoder checks removed after removing the class.
  2013/09/04 - Colored status in langhowto. *ALMOST up-to-date* category
               of translators introduced.
  2014/06/16 - unified for Python 2.6+ and 3.0+
  """

from __future__ import print_function

import os
import re
import sys
import textwrap


def xopen(fname, mode='r', encoding='utf-8-sig'):
    '''Unified file opening for Python 2 an Python 3.

    Python 2 does not have the encoding argument. Python 3 has one, and
    the default 'utf-8-sig' is used (skips the BOM automatically).
    '''

    if sys.version_info[0] == 2:
        return open(fname, mode=mode) # Python 2 without encoding
    else:
        return open(fname, mode=mode, encoding=encoding) # Python 3 with encoding


def fill(s):
    """Returns string formatted to the wrapped paragraph multiline string.

    Replaces whitespaces by one space and then uses he textwrap.fill()."""

    # Replace all whitespace by spaces, remove whitespaces that are not
    # necessary, strip the left and right whitespaces, and break the string
    # to list of words.
    rexWS = re.compile(r'\s+')
    lst = rexWS.sub(' ', s).strip().split()

    # If the list is not empty, put the words together and form the lines
    # of maximum 70 characters. Build the list of lines.
    lines = []
    if lst:
        line = lst.pop(0)   # no separation space in front of the first word
        for word in lst:
            if len(line) + len(word) < 70:
                line += ' ' + word
            else:
                lines.append(line)  # another full line formed
                line = word         # next line started
        lines.append(line)          # the last line
    return '\n'.join(lines)


class Transl:
    """One instance is build for each translator.

    The abbreviation of the source file--part after 'translator_'--is used as
    the identification of the object. The empty string is used for the
    abstract Translator class from translator.h. The other information is
    extracted from inside the source file."""

    def __init__(self, fname, manager):
        """Bind to the manager and initialize."""

        # Store the filename and the reference to the manager object.
        self.fname = fname
        self.manager = manager

        # The instance is responsible for loading the source file, so it checks
        # for its existence and quits if something goes wrong.
        if not os.path.isfile(fname):
            sys.stderr.write("\a\nFile '%s' not found!\n" % fname)
            sys.exit(1)

        # Initialize the other collected information.
        self.classId = None
        self.baseClassId = None
        self.readableStatus = None   # 'up-to-date', '1.2.3', '1.3', etc.
        self.status = None           # '', '1.2.03', '1.3.00', etc.
        self.lang = None             # like 'Brazilian'
        self.langReadable = None     # like 'Brazilian Portuguese'
        self.note = None             # like 'should be cleaned up'
        self.prototypeDic = {}       # uniPrototype -> prototype
        self.translateMeText = 'translate me!'
        self.translateMeFlag = False # comments with "translate me!" found
        self.txtMAX_DOT_GRAPH_HEIGHT_flag = False # found in string in trLegendDocs()
        self.obsoleteMethods = None  # list of prototypes to be removed
        self.missingMethods = None   # list of prototypes to be implemented
        self.implementedMethods = None  # list of implemented required methods
        self.adaptMinClass = None    # The newest adapter class that can be used

    def __tokenGenerator(self):
        """Generator that reads the file and yields tokens as 4-tuples.

        The tokens have the form (tokenId, tokenString, lineNo). The
        last returned token has the form ('eof', None, None). When trying
        to access next token after that, the exception would be raised."""

        # Set the dictionary for recognizing tokenId for keywords, separators
        # and the similar categories. The key is the string to be recognized,
        # the value says its token identification.
        tokenDic = { 'class':     'class',
                     'const':     'const',
                     'public':    'public',
                     'protected': 'protected',
                     'private':   'private',
                     'static':    'static',
                     'virtual':   'virtual',
                     'override':  'override',
                     ':':         'colon',
                     ';':         'semic',
                     ',':         'comma',
                     '[':         'lsqbra',
                     ']':         'rsqbra',
                     '(':         'lpar',
                     ')':         'rpar',
                     '{':         'lcurly',
                     '}':         'rcurly',
                     '=':         'assign',
                     '*':         'star',
                     '&':         'amp',
                     '+':         'plus',
                     '-':         'minus',
                     '!':         'excl',
                     '?':         'qmark',
                     '<':         'lt',
                     '>':         'gt',
                     "'":         'quot',
                     '"':         'dquot',
                     '.':         'dot',
                     '%':         'perc',
                     '~':         'tilde',
                     '^':         'caret',
                     '|':         'pipe',
                   }

        # Regular expression for recognizing identifiers.
        rexId = re.compile(r'^[a-zA-Z]\w*$')

        # Open the file for reading and extracting tokens until the eof.
        # Initialize the finite automaton.
        f = xopen(self.fname)
        lineNo = 0
        line = ''         # init -- see the pos initialization below
        linelen = 0       # init
        pos = 100         # init -- pos after the end of line
        status = 0

        tokenId = None    # init
        tokenStr = ''     # init -- the characters will be appended.
        tokenLineNo = 0

        while status != 777:

            # Get the next character. Read next line first, if necessary.
            if pos < linelen:
                c = line[pos]
            else:
                lineNo += 1
                line = f.readline()
                linelen = len(line)
                pos = 0
                if line == '':         # eof
                    status = 777
                else:
                    c = line[pos]

            # Consume the character based on the status

            if status == 0:     # basic status

                # This is the initial status. If tokenId is set, yield the
                # token here and only here (except when eof is found).
                # Initialize the token variables after the yield.
                if tokenId:
                    # If it is an unknown item, it can still be recognized
                    # here. Keywords and separators are the example.
                    if tokenId == 'unknown':
                        if tokenStr in tokenDic:
                            tokenId = tokenDic[tokenStr]
                        elif tokenStr.isdigit():
                            tokenId = 'num'
                        elif rexId.match(tokenStr):
                            tokenId = 'id'
                        else:
                            self.__unexpectedToken(-1, tokenStr, tokenLineNo)

                    yield (tokenId, tokenStr, tokenLineNo)

                    # If it is a comment that contains the self.translateMeText
                    # string, set the flag -- the situation will be reported.
                    if tokenId == 'comment' and tokenStr.find(self.translateMeText) >= 0:
                        self.translateMeFlag = True

                    tokenId = None
                    tokenStr = ''
                    tokenLineNo = 0

                # Now process the character. When we just skip it (spaces),
                # stay in this status. All characters that will be part of
                # some token cause moving to the specific status. And only
                # when moving to the status == 0 (or the final state 777),
                # the token is yielded. With respect to that the automaton
                # behaves as Moore's one (output bound to status). When
                # collecting tokens, the automaton is the Mealy's one
                # (actions bound to transitions).
                if c.isspace():
                    pass                 # just skip whitespace characters
                elif c == '/':           # Possibly comment starts here, but
                    tokenId = 'unknown'  # it could be only a slash in code.
                    tokenStr = c
                    tokenLineNo = lineNo
                    status = 1
                elif c == '#':
                    tokenId = 'preproc'  # preprocessor directive
                    tokenStr = c
                    tokenLineNo = lineNo
                    status = 5
                elif c == '"':           # string starts here
                    tokenId = 'string'
                    tokenStr = c
                    tokenLineNo = lineNo
                    status = 6
                elif c == "'":           # char literal starts here
                    tokenId = 'charlit'
                    tokenStr = c
                    tokenLineNo = lineNo
                    status = 8
                elif c in tokenDic:  # known one-char token
                    tokenId = tokenDic[c]
                    tokenStr = c
                    tokenLineNo = lineNo
                    # stay in this state to yield token immediately
                else:
                    tokenId = 'unknown'  # totally unknown
                    tokenStr = c
                    tokenLineNo = lineNo
                    status = 333

                pos += 1                 # move position in any case

            elif status == 1:            # possibly a comment
                if c == '/':             # ... definitely the C++ comment
                    tokenId = 'comment'
                    tokenStr += c
                    pos += 1
                    status = 2
                elif c == '*':           # ... definitely the C comment
                    tokenId = 'comment'
                    tokenStr += c
                    pos += 1
                    status = 3
                else:
                    status = 0           # unrecognized, don't move pos

            elif status == 2:            # inside the C++ comment
                if c == '\n':            # the end of C++ comment
                    status = 0           # yield the token
                else:
                    tokenStr += c        # collect the C++ comment
                pos += 1

            elif status == 3:            # inside the C comment
                if c == '*':             # possibly the end of the C comment
                    tokenStr += c
                    status = 4
                else:
                    tokenStr += c        # collect the C comment
                pos += 1

            elif status == 4:            # possibly the end of the C comment
                if c == '/':             # definitely the end of the C comment
                    tokenStr += c
                    status = 0           # yield the token
                elif c == '*':           # more stars inside the comment
                    tokenStr += c
                else:
                    tokenStr += c        # this cannot be the end of comment
                    status = 3
                pos += 1

            elif status == 5:            # inside the preprocessor directive
                if c == '\n':            # the end of the preproc. command
                    status = 0           # yield the token
                else:
                    tokenStr += c        # collect the preproc
                pos += 1

            elif status == 6:            # inside the string
                if c == '\\':            # escaped char inside the string
                    tokenStr += c
                    status = 7
                elif c == '"':           # end of the string
                    tokenStr += c
                    status = 0
                else:
                    tokenStr += c        # collect the chars of the string
                pos += 1

            elif status == 7:            # escaped char inside the string
                tokenStr += c            # collect the char of the string
                status = 6
                pos += 1

            elif status == 8:            # inside the char literal
                tokenStr += c            # collect the char of the literal
                status = 9
                pos += 1

            elif status == 9:            # end of char literal expected
                if c == "'":             # ... and found
                    tokenStr += c
                    status = 0
                    pos += 1
                else:
                    tokenId = 'error'    # end of literal was expected
                    tokenStr += c
                    status = 0

            elif status == 333:          # start of the unknown token
                if c.isspace():
                    pos += 1
                    status = 0           # tokenId may be determined later
                elif c in tokenDic:  # separator, don't move pos
                    status = 0
                else:
                    tokenStr += c        # collect
                    pos += 1

        # We should have finished in the final status. If some token
        # have been extracted, yield it first.
        assert(status == 777)
        if tokenId:
            yield (tokenId, tokenStr, tokenLineNo)
            tokenId = None
            tokenStr = ''
            tokenLineNo = 0

        # The file content is processed. Close the file. Then always yield
        # the eof token.
        f.close()
        yield ('eof', None, None)


    def __collectClassInfo(self, tokenIterator):
        """Collect the information about the class and base class.

        The tokens including the opening left curly brace of the class are
        consumed."""

        status = 0  # initial state

        while status != 777:   # final state

            # Always assume that the previous tokens were processed. Get
            # the next one.
            tokenId, tokenStr, tokenLineNo = next(tokenIterator)

            # Process the token and never return back.
            if status == 0:    # waiting for the 'class' keyword.
                if tokenId == 'class':
                    status = 1

            elif status == 1:  # expecting the class identification
                if tokenId == 'id':
                    self.classId = tokenStr
                    status = 2
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 2:  # expecting the curly brace or base class info
                if tokenId == 'lcurly':
                    status = 777        # correctly finished
                elif tokenId == 'colon':
                    status = 3
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 3:  # expecting the 'public' in front of base class id
                if tokenId == 'public':
                    status = 4
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 4:  # expecting the base class id
                if tokenId == 'id':
                    self.baseClassId = tokenStr
                    status = 5
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 5:  # expecting the curly brace and quitting
                if tokenId == 'lcurly':
                    status = 777        # correctly finished
                elif tokenId == 'comment':
                    pass
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

        # Extract the status of the TranslatorXxxx class. The readable form
        # will be used in reports the status form is a string that can be
        # compared lexically (unified length, padding with zeros, etc.).
        if self.baseClassId:
            lst = self.baseClassId.split('_')
            if lst[0] == 'Translator':
                self.readableStatus = 'up-to-date'
                self.status = ''
            elif lst[0] == 'TranslatorAdapter':
                self.status = lst[1] + '.' + lst[2]
                self.readableStatus = self.status
                if len(lst) > 3:        # add the last part of the number
                    self.status += '.' + ('%02d' % int(lst[3]))
                    self.readableStatus += '.' + lst[3]
                else:
                    self.status += '.00'
            elif lst[0] == 'TranslatorEnglish':
                # Obsolete or Based on English.
                if self.classId[-2:] == 'En':
                    self.readableStatus = 'English based'
                    self.status = 'En'
                else:
                    self.readableStatus = 'obsolete'
                    self.status = '0.0.00'

            # Check whether status was set, or set 'strange'.
            if self.status is None:
                self.status = 'strange'
            if not self.readableStatus:
                self.readableStatus = 'strange'

            # Extract the name of the language and the readable form.
            self.lang = self.classId[10:]  # without 'Translator'
            if self.lang == 'Brazilian':
                self.langReadable = 'Brazilian Portuguese'
            elif self.lang == 'Chinesetraditional':
                self.langReadable = 'Chinese Traditional'
            else:
                self.langReadable = self.lang


    def __unexpectedToken(self, status, tokenId, tokenLineNo):
        """Reports unexpected token and quits with exit code 1."""

        import inspect
        calledFrom = inspect.stack()[1][3]
        msg = "\a\nUnexpected token '%s' on the line %d in '%s'.\n"
        msg = msg % (tokenId, tokenLineNo, self.fname)
        if status != -1:
            msg += 'status = %d in %s()\n' % (status, calledFrom)
        sys.stderr.write(msg)
        sys.exit(1)


    def collectPureVirtualPrototypes(self):
        """Returns dictionary 'unified prototype' -> 'full prototype'.

        The method is expected to be called only for the translator.h. It
        extracts only the pure virtual method and build the dictionary where
        key is the unified prototype without argument identifiers."""

        # Prepare empty dictionary that will be returned.
        resultDic = {}

        # Start the token generator which parses the class source file.
        tokenIterator = self.__tokenGenerator()

        # Collect the class and the base class identifiers.
        self.__collectClassInfo(tokenIterator)
        assert(self.classId == 'Translator')

        # Let's collect readable form of the public virtual pure method
        # prototypes in the readable form -- as defined in translator.h.
        # Let's collect also unified form of the same prototype that omits
        # everything that can be omitted, namely 'virtual' and argument
        # identifiers.
        prototype = ''    # readable prototype (with everything)
        uniPrototype = '' # unified prototype (without arg. identifiers)

        # Collect the pure virtual method prototypes. Stop on the closing
        # curly brace followed by the semicolon (end of class).
        status = 0
        curlyCnt = 0      # counter for the level of curly braces

        backStatus = 2
        # Loop until the final state 777 is reached. The errors are processed
        # immediately. In this implementation, it always quits the application.
        while status != 777:

            # Get the next token.
            tokenId, tokenStr, tokenLineNo = next(tokenIterator)

            if status == 0:      # waiting for 'public:'
                if tokenId == 'public':
                    status = 1

            elif status == 1:    # colon after the 'public'
                if tokenId == 'colon':
                    backStatus = 2
                    status = 2
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 2:    # waiting for 'virtual'
                if tokenId == 'virtual':
                    prototype = tokenStr  # but not to unified prototype
                    status = 3
                elif tokenId == 'id' and tokenStr == 'QCString' and backStatus == 3:
                    status = 4
                elif tokenId == 'comment':
                    pass
                elif tokenId == 'rcurly':
                    status = 11         # expected end of class
                elif tokenId == 'id' and tokenStr == 'ABSTRACT_BASE_CLASS':
                    status = 18
                elif tokenId == 'protected':
                    status = 19
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 3:    # return type of the method expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    uniPrototype = tokenStr  # start collecting the unified prototype
                    status = 4
                elif tokenId == 'tilde':
                    status = 4
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 4:    # method identifier expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    uniPrototype += ' ' + tokenStr
                    status = 5
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 5:    # left bracket of the argument list expected
                if tokenId == 'lpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 6
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 6:    # collecting arguments of the method
                if tokenId == 'rpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 7
                elif tokenId == 'const':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 12
                elif tokenId == 'id':           # type identifier
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 13
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 7:    # assignment expected or left curly brace
                if tokenId == 'assign':
                    status = 8
                elif tokenId == 'lcurly':
                    curlyCnt = 1      # method body entered
                    status = 10
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 8:    # zero expected (or default for the destructor)
                if (tokenId == 'num' and tokenStr == '0') or (tokenId == 'id' and tokenStr == 'default'):
                    status = 9
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 9:    # after semicolon, produce the dic item
                if tokenId == 'semic':
                    assert(uniPrototype not in resultDic)
                    resultDic[uniPrototype] = prototype
                    status = 2
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 10:   # consuming the body of the method
                if tokenId == 'rcurly':
                    curlyCnt -= 1
                    if curlyCnt == 0:
                        status = 2     # body consumed
                elif tokenId == 'lcurly':
                    curlyCnt += 1

            elif status == 11:   # probably the end of class
                if tokenId == 'semic':
                    status = 777
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 12:   # type id for argument expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    uniPrototype += ' ' + tokenStr
                    status = 13
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 13:   # namespace qualification or * or & expected
                if tokenId == 'colon':        # was namespace id
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 14
                elif tokenId == 'star' or tokenId == 'amp':  # pointer or reference
                    prototype += ' ' + tokenStr
                    uniPrototype += ' ' + tokenStr
                    status = 16
                elif tokenId == 'id':         # argument identifier
                    prototype += ' ' + tokenStr
                    # don't put this into unified prototype
                    status = 17
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 14:   # second colon for namespace:: expected
                if tokenId == 'colon':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 15
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 15:   # type after namespace:: expected
                if tokenId == 'id':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 13
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 16:   # argument identifier expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    # don't put this into unified prototype
                    status = 17
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 17:   # comma or ')' after argument identifier expected
                if tokenId == 'comma':
                    prototype += ', '
                    uniPrototype += ', '
                    status = 6
                elif tokenId == 'assign':
                    status=20
                elif tokenId == 'rpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 7
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 18:    # start of the ABSTRACT_BASE_CLASS
                if tokenId == 'lpar':
                    pass
                elif tokenId == 'rpar':
                    status = 2
                elif tokenId == 'id':
                    pass
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 19:    # colon after the 'protected'
                if tokenId == 'colon':
                    backStatus = 3
                    status = 3
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 20:
                if tokenId == 'string':
                    status = 17
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

        # Eat the rest of the source to cause closing the file.
        while tokenId != 'eof':
            tokenId, tokenStr, tokenLineNo = next(tokenIterator)

        # Return the resulting dictionary with 'uniPrototype -> prototype'.
        return resultDic


    def __collectPublicMethodPrototypes(self, tokenIterator):
        """Collects prototypes of public methods and fills self.prototypeDic.

        The dictionary is filled by items: uniPrototype -> prototype.
        The method is expected to be called only for TranslatorXxxx classes,
        i.e. for the classes that implement translation to some language.
        It assumes that the opening curly brace of the class was already
        consumed. The source is consumed until the end of the class.
        The caller should consume the source until the eof to cause closing
        the source file."""

        assert(self.classId != 'Translator')
        assert self.baseClassId is not None, 'Class ' + self.classId + ' from the file ' + self.fname + ' does not have a base class.'

        # The following finite automaton slightly differs from the one
        # inside self.collectPureVirtualPrototypes(). It produces the
        # dictionary item just after consuming the body of the method
        # (transition from state 10 to state 2). It also does not allow
        # definitions of public pure virtual methods, except for
        # TranslatorAdapterBase (states 8 and 9). Argument identifier inside
        # method argument lists can be omitted or commented.
        #
        # Let's collect readable form of all public method prototypes in
        # the readable form -- as defined in the source file.
        # Let's collect also unified form of the same prototype that omits
        # everything that can be omitted, namely 'virtual' and argument
        # identifiers.
        prototype = ''    # readable prototype (with everything)
        uniPrototype = '' # unified prototype (without arg. identifiers)
        methodId = None   # processed method id

        # Collect the method prototypes. Stop on the closing
        # curly brace followed by the semicolon (end of class).
        status = 0
        curlyCnt = 0      # counter for the level of curly braces

        # Loop until the final state 777 is reached. The errors are processed
        # immediately. In this implementation, it always quits the application.
        while status != 777:

            # Get the next token.
            tokenId, tokenStr, tokenLineNo = next(tokenIterator)

            if status == 0:      # waiting for 'public:'
                if tokenId == 'public':
                    status = 1
                elif tokenId == 'eof':  # non-public things until the eof
                    status = 777

            elif status == 1:    # colon after the 'public'
                if tokenId == 'colon':
                    status = 2
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 2:    # waiting for 'virtual' (can be omitted)
                if tokenId == 'virtual':
                    prototype = tokenStr  # but not to unified prototype
                    status = 3
                elif tokenId == 'id':     # 'virtual' was omitted
                    prototype = tokenStr
                    uniPrototype = tokenStr  # start collecting the unified prototype
                    status = 4
                elif tokenId == 'comment':
                    pass
                elif tokenId == 'protected' or tokenId == 'private':
                    status = 0
                elif tokenId == 'rcurly':
                    status = 11         # expected end of class
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 3:    # return type of the method expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    uniPrototype = tokenStr  # start collecting the unified prototype
                    status = 4
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 4:    # method identifier expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    uniPrototype += ' ' + tokenStr
                    methodId = tokenStr    # for reporting
                    status = 5
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 5:    # left bracket of the argument list expected
                if tokenId == 'lpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 6
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 6:    # collecting arguments of the method
                if tokenId == 'rpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 7
                elif tokenId == 'const':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 12
                elif tokenId == 'id':           # type identifier
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 13
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 7:    # left curly brace expected
                if tokenId == 'lcurly':
                    curlyCnt = 1      # method body entered
                    status = 10
                elif tokenId == 'override':
                    pass
                elif tokenId == 'comment':
                    pass
                elif tokenId == 'assign': # allowed only for TranslatorAdapterBase
                    assert(self.classId == 'TranslatorAdapterBase')
                    status = 8
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 8:    # zero expected (TranslatorAdapterBase)
                assert(self.classId == 'TranslatorAdapterBase')
                if tokenId == 'num' and tokenStr == '0':
                    status = 9
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 9:    # after semicolon (TranslatorAdapterBase)
                assert(self.classId == 'TranslatorAdapterBase')
                if tokenId == 'semic':
                    status = 2
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 10:   # consuming the body of the method, then dic item
                if tokenId == 'rcurly':
                    curlyCnt -= 1
                    if curlyCnt == 0:
                        # Check for possible copy/paste error when name
                        # of the method was not corrected (i.e. the same
                        # name already exists).
                        if uniPrototype in self.prototypeDic:
                            msg = "'%s' prototype found again (duplicity)\n"
                            msg += "in '%s'.\n" % self.fname
                            msg = msg % uniPrototype
                            sys.stderr.write(msg)
                            assert False

                        assert(uniPrototype not in self.prototypeDic)
                        # Insert new dictionary item, unless they have a default in translator.h
                        if (not (prototype=="QCString latexDocumentPost()" or
                                 prototype=="QCString latexDocumentPre()" or
                                 prototype=="QCString latexCommandName()" or
                                 prototype=="QCString latexFont()" or
                                 prototype=="QCString latexFontenc()" or
                                 prototype=="bool needsPunctuation()")):
                            self.prototypeDic[uniPrototype] = prototype
                        status = 2      # body consumed
                        methodId = None # outside of any method
                elif tokenId == 'lcurly':
                    curlyCnt += 1

                # Warn in special case.
                elif methodId == 'trLegendDocs' and tokenId == 'string' \
                    and tokenStr.find('MAX_DOT_GRAPH_HEIGHT') >= 0:
                        self.txtMAX_DOT_GRAPH_HEIGHT_flag = True


            elif status == 11:   # probably the end of class
                if tokenId == 'semic':
                    status = 777
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 12:   # type id for argument expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    uniPrototype += ' ' + tokenStr
                    status = 13
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 13:   # :: or * or & or id or ) expected
                if tokenId == 'colon':        # was namespace id
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 14
                elif tokenId == 'star' or tokenId == 'amp':  # pointer or reference
                    prototype += ' ' + tokenStr
                    uniPrototype += ' ' + tokenStr
                    status = 16
                elif tokenId == 'id':         # argument identifier
                    prototype += ' ' + tokenStr
                    # don't put this into unified prototype
                    status = 17
                elif tokenId == 'comment':    # probably commented-out identifier
                    prototype += tokenStr
                elif tokenId == 'rpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 7
                elif tokenId == 'comma':
                    prototype += ', '
                    uniPrototype += ', '
                    status = 6
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 14:   # second colon for namespace:: expected
                if tokenId == 'colon':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 15
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 15:   # type after namespace:: expected
                if tokenId == 'id':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 13
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 16:   # argument identifier or ) expected
                if tokenId == 'id':
                    prototype += ' ' + tokenStr
                    # don't put this into unified prototype
                    status = 17
                elif tokenId == 'rpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 7
                elif tokenId == 'comment':
                    prototype += tokenStr
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)

            elif status == 17:   # comma or ')' after argument identifier expected
                if tokenId == 'comma':
                    prototype += ', '
                    uniPrototype += ', '
                    status = 6
                elif tokenId == 'rpar':
                    prototype += tokenStr
                    uniPrototype += tokenStr
                    status = 7
                else:
                    self.__unexpectedToken(status, tokenId, tokenLineNo)



    def collectAdapterPrototypes(self):
        """Returns the dictionary of prototypes implemented by adapters.

        It is created to process the translator_adapter.h. The returned
        dictionary has the form: unifiedPrototype -> (version, classId)
        thus by looking for the prototype, we get the information what is
        the newest (least adapting) adapter that is sufficient for
        implementing the method."""

        # Start the token generator which parses the class source file.
        assert(os.path.split(self.fname)[1] == 'translator_adapter.h')
        tokenIterator = self.__tokenGenerator()

        # Get the references to the involved dictionaries.
        reqDic = self.manager.requiredMethodsDic

        # Create the empty dictionary that will be returned.
        adaptDic = {}


        # Loop through the source of the adapter file until no other adapter
        # class is found.
        while True:
            try:
                # Collect the class and the base class identifiers.
                self.__collectClassInfo(tokenIterator)

                # Extract the comparable version of the adapter class.
                # Note: The self.status as set by self.__collectClassInfo()
                # contains similar version, but is related to the base class,
                # not to the class itself.
                lst = self.classId.split('_')
                version = ''
                if lst[0] == 'TranslatorAdapter': # TranslatorAdapterBase otherwise
                    version = lst[1] + '.' + lst[2]
                    if len(lst) > 3:        # add the last part of the number
                        version += '.' + ('%02d' % int(lst[3]))
                    else:
                        version += '.00'

                # Collect the prototypes of implemented public methods.
                self.__collectPublicMethodPrototypes(tokenIterator)

                # For the required methods, update the dictionary of methods
                # implemented by the adapter.
                for protoUni in self.prototypeDic:
                    if protoUni in reqDic:
                        # This required method will be marked as implemented
                        # by this adapter class. This implementation assumes
                        # that newer adapters do not reimplement any required
                        # methods already implemented by older adapters.
                        assert(protoUni not in adaptDic)
                        adaptDic[protoUni] = (version, self.classId)

                # Clear the dictionary object and the information related
                # to the class as the next adapter class is to be processed.
                self.prototypeDic.clear()
                self.classId = None
                self.baseClassId = None

            except StopIteration:
                break

        # Return the result dictionary.
        return adaptDic


    def processing(self):
        """Processing of the source file -- only for TranslatorXxxx classes."""

        # Start the token generator which parses the class source file.
        tokenIterator = self.__tokenGenerator()

        # Collect the class and the base class identifiers.
        self.__collectClassInfo(tokenIterator)
        assert(self.classId != 'Translator')
        assert(self.classId[:17] != 'TranslatorAdapter')

        # Collect the prototypes of implemented public methods.
        self.__collectPublicMethodPrototypes(tokenIterator)

        # Eat the rest of the source to cause closing the file.
        while True:
            try:
                next(tokenIterator)
            except StopIteration:
                break

        # Shorthands for the used dictionaries.
        reqDic = self.manager.requiredMethodsDic
        adaptDic = self.manager.adaptMethodsDic
        myDic = self.prototypeDic

        # Build the list of obsolete methods.
        self.obsoleteMethods = []
        for p in myDic:
            if p not in reqDic:
                self.obsoleteMethods.append(p)
        self.obsoleteMethods.sort()

        # Build the list of missing methods and the list of implemented
        # required methods.
        self.missingMethods = []
        self.implementedMethods = []
        for p in reqDic:
            if p in myDic:
                self.implementedMethods.append(p)
            else:
                self.missingMethods.append(p)
        self.missingMethods.sort()
        self.implementedMethods.sort()

        # Check whether adapter must be used or suggest the newest one.
        # Change the status and set the note accordingly.
        if self.baseClassId != 'Translator':
            justUpdateNeesedMessage = True
            if not self.missingMethods:
                self.note = 'Change the base class to Translator.'
                self.status = ''
                self.readableStatus = 'almost up-to-date'
            elif self.baseClassId != 'TranslatorEnglish':
                # The translator uses some of the adapters.
                # Look at the missing methods and check what adapter
                # implements them. Remember the one with the lowest version.
                adaptMinVersion = '9.9.99'
                adaptMinClass = 'TranslatorAdapter_9_9_99'
                for uniProto in self.missingMethods:
                    if uniProto in adaptDic:
                        version, cls = adaptDic[uniProto]
                        if version < adaptMinVersion:
                            justUpdateNeesedMessage = False
                            adaptMinVersion = version
                            adaptMinClass = cls

                if justUpdateNeesedMessage:
                    self.note = 'Change the base class to Translator.'
                    self.status = ''

                # Test against the current status -- preserve the self.status.
                # Possibly, the translator implements enough methods to
                # use some newer adapter.
                status = self.status

                # If the version of the used adapter is smaller than
                # the required, set the note and update the status as if
                # the newer adapter was used.
                if not justUpdateNeesedMessage and adaptMinVersion > status:
                    self.note = 'Change the base class to %s.' % adaptMinClass
                    self.status = adaptMinVersion
                    self.adaptMinClass = adaptMinClass
                    self.readableStatus = adaptMinVersion # simplified

        # If everything seems OK, some explicit warning flags still could
        # be set.
        if not self.note and self.status == '' and \
           (self.translateMeFlag or self.txtMAX_DOT_GRAPH_HEIGHT_flag):
            self.note = ''
            if self.translateMeFlag:
                self.note += 'The "%s" found in a comment.' % self.translateMeText
            if self.note != '':
                self.note += '\n\t\t'
            if self.txtMAX_DOT_GRAPH_HEIGHT_flag:
                self.note += 'The MAX_DOT_GRAPH_HEIGHT found in trLegendDocs()'

        # If everything seems OK, but there are obsolete methods, set
        # the note to clean-up source. This note will be used only when
        # the previous code did not set another note (priority).
        if not self.note and self.status == '' and self.obsoleteMethods:
            self.note = 'Remove the obsolete methods (never used).'

        # If there is at least some note but the status suggests it is
        # otherwise up-to-date, mark is as ALMOST up-to-date.
        if self.note and self.status == '':
            self.readableStatus = 'almost up-to-date'


    def report(self, fout):
        """Returns the report part for the source as a multiline string.

        No output for up-to-date translators without problem."""

        # If there is nothing to report, return immediately.
        if self.status == '' and not self.note:
            return

        # Report the number of not implemented methods.
        fout.write('\n\n\n')
        fout.write(self.classId + '   (' + self.baseClassId + ')')
        percentImplemented = 100    # init
        allNum = len(self.manager.requiredMethodsDic)
        if self.missingMethods:
            num = len(self.missingMethods)
            percentImplemented = 100 * (allNum - num) / allNum
            fout.write('  %d' % num)
            fout.write(' method')
            if num > 1:
                fout.write('s')
            fout.write(' to implement (%d %%)' % (100 * num / allNum))
        fout.write('\n' + '-' * len(self.classId))

        # Write the info about the implemented required methods.
        fout.write('\n\n  Implements %d' % len(self.implementedMethods))
        fout.write(' of the required methods (%d %%).' % percentImplemented)

        # Report the missing method, but only when it is not English-based
        # translator.
        if self.missingMethods and self.status != 'En':
            fout.write('\n\n  Missing methods (should be implemented):\n')
            reqDic = self.manager.requiredMethodsDic
            for p in self.missingMethods:
                fout.write('\n    ' + reqDic[p])

        # Always report obsolete methods.
        if self.obsoleteMethods:
            fout.write('\n\n  Obsolete methods (should be removed, never used):\n')
            myDic = self.prototypeDic
            for p in self.obsoleteMethods:
                fout.write('\n    ' + myDic[p])

        # For English-based translator, report the implemented methods.
        if self.status == 'En' and self.implementedMethods:
            fout.write('\n\n  This English-based translator implements ')
            fout.write('the following methods:\n')
            reqDic = self.manager.requiredMethodsDic
            for p in self.implementedMethods:
                fout.write('\n    ' + reqDic[p])


    def getmtime(self):
        """Returns the last modification time of the source file."""
        assert(os.path.isfile(self.fname))
        return os.path.getmtime(self.fname)


class TrManager:
    """Collects basic info and builds subordinate Transl objects."""

    def __init__(self):
        """Determines paths, creates and initializes structures.

        The arguments of the script may explicitly say what languages should
        be processed. Write the two letter identifications that are used
        for composing the source filenames, so...

            python translator.py cz

        this will process only translator_cz.h source.
        """

        # Determine the path to the script and its name.
        self.script = os.path.abspath(sys.argv[0])
        self.script_path, self.script_name = os.path.split(self.script)
        self.script_path = os.path.abspath(self.script_path)

        # Determine the absolute path to the Doxygen's root subdirectory.
        # If DOXYGEN environment variable is not found, the directory is
        # determined from the path of the script.
        doxy_default = os.path.join(self.script_path, '..')
        self.doxy_path = os.path.abspath(os.getenv('DOXYGEN', doxy_default))

        self.internal = False
        if sys.argv[1] == '--doc':
            self.internal = False
        elif sys.argv[1] == '--doc_internal':
            self.internal = True

        # Build the path names based on the Doxygen's root knowledge.
        if self.internal:
            self.doc_path = os.path.join(self.doxy_path, 'doc_internal')
        else:
            self.doc_path = os.path.join(self.doxy_path, 'doc')
        self.src_path = os.path.join(self.doxy_path, 'src')
        #  Normally the original sources aren't in the current directory
        # (as we are in the build directory) so we have to specify the
        # original source /documentation / ... directory.
        self.org_src_path = os.path.join(sys.argv[2], 'src')
        self.org_doc_path = os.path.join(sys.argv[2], 'doc')
        self.org_doxy_path = sys.argv[2]

        # Create the empty dictionary for Transl object identified by the
        # class identifier of the translator.
        self.__translDic = {}

        # Create the None dictionary of required methods. The key is the
        # unified prototype, the value is the full prototype. Set inside
        # the self.__build().
        self.requiredMethodsDic = None

        # Create the empty dictionary that says what method is implemented
        # by what adapter.
        self.adaptMethodsDic = {}

        # The last modification time will capture the modification of this
        # script, of the translator.h, of the translator_adapter.h (see the
        # self.__build() for the last two) of all the translator_xx.h files
        # and of the template for generating the documentation. So, this
        # time can be compared with modification time of the generated
        # documentation to decide, whether the doc should be re-generated.
        self.lastModificationTime = os.path.getmtime(self.script)

        # Set the names of the translator report text file, of the template
        # for generating "Internationalization" document, for the generated
        # file itself, and for the maintainers list.

        if self.internal:
            self.translatorReportFileName = 'translator_report.md'
        else:
            self.translatorReportFileName = 'translator_report.txt'
        self.maintainersFileName = 'maintainers.txt'
        self.languageTplFileName = 'language.tpl'
        self.languageDocFileName = 'language.dox'

        # The information about the maintainers will be stored
        # in the dictionary with the following name.
        self.__maintainersDic = None

        # Define the other used structures and variables for information.
        self.langLst = None                   # including English based
        self.supportedLangReadableStr = None  # coupled En-based as a note
        self.numLang = None                   # excluding coupled En-based
        self.doxVersion = None                # Doxygen version

        # Build objects where each one is responsible for one translator.
        self.__build()


    def __build(self):
        """Find the translator files and build the objects for translators."""

        # The translator.h must exist (the Transl object will check it),
        # create the object for it and let it build the dictionary of
        # required methods.
        tr = Transl(os.path.join(self.org_src_path, 'translator.h'), self)
        self.requiredMethodsDic = tr.collectPureVirtualPrototypes()
        tim = tr.getmtime()
        if tim > self.lastModificationTime:
            self.lastModificationTime = tim

        # The translator_adapter.h must exist (the Transl object will check it),
        # create the object for it and store the reference in the dictionary.
        tr = Transl(os.path.join(self.org_src_path, 'translator_adapter.h'), self)
        self.adaptMethodsDic = tr.collectAdapterPrototypes()
        tim = tr.getmtime()
        if tim > self.lastModificationTime:
            self.lastModificationTime = tim

        # Create the list of the filenames with language translator sources.
        lst = os.listdir(self.org_src_path)
        lst = [x for x in lst if x[:11] == 'translator_'
                               and x[-2:] == '.h'
                               and x != 'translator_adapter.h']

        # Build the object for the translator_xx.h files, and process the
        # content of the file. Then insert the object to the dictionary
        # accessed via classId.
        for fname in lst:
            fullname = os.path.join(self.org_src_path, fname)
            tr = Transl(fullname, self)
            tr.processing()
            assert(tr.classId != 'Translator')
            self.__translDic[tr.classId] = tr

        # Extract the global information of the processed info.
        self.__extractProcessedInfo()


    def __extractProcessedInfo(self):
        """Build lists and strings of the processed info."""

        # Build the auxiliary list with strings compound of the status,
        # readable form of the language, and classId.
        statLst = []
        for obj in list(self.__translDic.values()):
            assert(obj.classId != 'Translator')
            s = obj.status + '|' + obj.langReadable + '|' + obj.classId
            statLst.append(s)

        # Sort the list and extract the object identifiers (classId's) for
        # the up-to-date translators and English-based translators.
        statLst.sort()
        self.upToDateIdLst = [x.split('|')[2] for x in statLst if x[0] == '|']
        self.EnBasedIdLst = [x.split('|')[2] for x in statLst if x[:2] == 'En']

        # Reverse the list and extract the TranslatorAdapter based translators.
        statLst.reverse()
        self.adaptIdLst = [x.split('|')[2] for x in statLst if x[0].isdigit()]

        # Build the list of tuples that contain (langReadable, obj).
        # Sort it by readable name.
        self.langLst = []
        for obj in list(self.__translDic.values()):
            self.langLst.append((obj.langReadable, obj))

        self.langLst.sort(key=lambda x: x[0])

        # Create the list with readable language names. If the language has
        # also the English-based version, modify the item by appending
        # the note. Number of the supported languages is equal to the length
        # of the list.
        langReadableLst = []
        for name, obj in self.langLst:
            if obj.status == 'En':
                continue

            # Append the 'En' to the classId to possibly obtain the classId
            # of the English-based object. If the object exists, modify the
            # name for the readable list of supported languages.
            classIdEn = obj.classId + 'En'
            if classIdEn in self.__translDic:
                name += ' (+En)'

            # Append the result name of the language, possibly with note.
            langReadableLst.append(name)

        # Create the multiline string of readable language names,
        # with punctuation, wrapped to paragraph.
        if len(langReadableLst) == 1:
            s = langReadableLst[0]
        elif len(langReadableLst) == 2:
            s = ' and '.join(langReadableLst)
        else:
            s = ', '.join(langReadableLst[:-1]) + ', and '
            s += langReadableLst[-1]

        self.supportedLangReadableStr = fill(s + '.')

        # Find the number of the supported languages. The English based
        # languages are not counted if the non-English based also exists.
        self.numLang = len(self.langLst)
        for name, obj in self.langLst:
            if obj.status == 'En':
                classId = obj.classId[:-2]
                if classId in self.__translDic:
                    self.numLang -= 1    # the couple will be counted as one

        # Extract the version of Doxygen.
        f = xopen(os.path.join(self.org_doxy_path, 'VERSION'))
        self.doxVersion = f.readline().strip()
        f.close()

        # Update the last modification time.
        for tr in list(self.__translDic.values()):
            tim = tr.getmtime()
            if tim > self.lastModificationTime:
                self.lastModificationTime = tim


    def __getNoTrSourceFilesLst(self):
        """Returns the list of sources to be checked.

        All .cpp files and also .h files that do not declare or define
        the translator methods are included in the list. The file names
        are searched in doxygen/src directory.
        """
        files = []
        for item in os.listdir(self.org_src_path):
            # Split the bare name to get the extension.
            name, ext = os.path.splitext(item)
            ext = ext.lower()

            # Include only .cpp and .h files (case independent) and exclude
            # the files where the checked identifiers are defined.
            if ext == '.cpp' or ext ==  '.l' or (ext == '.h' and name.find('translator') == -1):
                fname = os.path.join(self.org_src_path, item)
                assert os.path.isfile(fname) # assumes no directory with the ext
                files.append(fname)          # full name
        return files


    def __removeUsedInFiles(self, fname, dic):
        """Removes items for method identifiers that are found in fname.

        The method reads the content of the file as one string and searches
        for all identifiers from dic. The identifiers that were found in
        the file are removed from the dictionary.

        Note: If more files is to be checked, the files where most items are
        probably used should be checked first and the resulting reduced
        dictionary should be used for checking the next files (speed up).
        """
        lst_in = list(dic.keys())   # identifiers to be searched for

        # Read content of the file as one string.
        assert os.path.isfile(fname)
        f = xopen(fname)
        cont = f.read()
        cont = ''.join(cont.split('\n')) # otherwise the 'match' function won't work.
        f.close()

        # Remove the items for identifiers that were found in the file.
        while lst_in:
            item = lst_in.pop(0)
            rexItem = re.compile(r'.*' + item + r' *\(')
            if rexItem.match(cont):
                del dic[item]


    def __checkForNotUsedTrMethods(self):
        """Returns the dictionary of not used translator methods.

        The method can be called only after self.requiredMethodsDic has been
        built. The stripped prototypes are the values, the method identifiers
        are the keys.
        """
        # Build the dictionary of the required method prototypes with
        # method identifiers used as keys.
        trdic = {}
        for prototype in list(self.requiredMethodsDic.keys()):
            ri = prototype.split('(')[0]
            identifier = ri.split()[1].strip()
            trdic[identifier] = prototype

        # Build the list of source files where translator method identifiers
        # can be used.
        files = self.__getNoTrSourceFilesLst()

        # Loop through the files and reduce the dictionary of id -> proto.
        for fname in files:
            self.__removeUsedInFiles(fname, trdic)

        # Return the dictionary of not used translator methods.
        return trdic


    def __emails(self, classId):
        """Returns the list of maintainer emails.

        The method returns the list of e-mail addresses for the translator
        class, but only the addresses that were not marked as [xxx]."""
        lst = []
        for m in self.__maintainersDic[classId]:
            if not m[1].startswith('['):
                email = m[1]
                email = email.replace(' at ', '@') # Unmangle the mangled e-mail
                email = email.replace(' dot ', '.')
                lst.append(email)
        return lst


    def getBgcolorByReadableStatus(self, readableStatus):
        if readableStatus == 'up-to-date':
            color = '#ccffcc'    # green
        elif readableStatus.startswith('almost'):
            color = '#ffffff'    # white
        elif readableStatus.startswith('English'):
            color = '#ccffcc'    # green
        elif readableStatus.startswith('1.9'):
            color = '#ffffcc'    # yellow
        elif readableStatus.startswith('1.8'):
            color = '#ffcccc'    # pink
        elif readableStatus.startswith('1.7'):
            color = '#ff5555'    # red
        elif readableStatus.startswith('1.6'):
            color = '#ff5555'    # red
        else:
            color = '#ff5555'    # red
        return color


    def generateTranslatorReport(self):
        """Generates the translator report."""

        output = os.path.join(self.doc_path, self.translatorReportFileName)

        # Open the textual report file for the output.
        f = xopen(output, 'w')

        # Output the information about the version.
        if self.internal:
            f.write('@page pg_trans Translator report\n\n')
            f.write('@verbatim\n\n')
        else:
            f.write('(' + self.doxVersion + ')\n\n')

        # Output the information about the number of the supported languages
        # and the list of the languages.
        f.write('Doxygen supports the following ')
        f.write(str(self.numLang))
        f.write(' languages (sorted alphabetically):\n\n')
        f.write(self.supportedLangReadableStr + '\n\n')

        # Write the summary about the status of language translators (how
        # many translators) are up-to-date, etc.
        s = 'Of them, %d translators are up-to-date, ' % len(self.upToDateIdLst)
        s += '%d translators are based on some adapter class, ' % len(self.adaptIdLst)
        s += 'and %d are English based.' % len(self.EnBasedIdLst)
        f.write(fill(s) + '\n\n')

        # The e-mail addresses of the maintainers will be collected to
        # the auxiliary file in the order of translator classes listed
        # in the translator report.
        fmail = xopen(os.path.join(self.doc_path, 'mailto.txt'), 'w')

        # Write the list of "up-to-date" translator classes.
        if self.upToDateIdLst:
            s = '''The following translator classes are up-to-date (sorted
                alphabetically). This means that they derive from the
                Translator class, they implement all %d of the required
                methods, and even minor problems were not spotted by the script:'''
            s = s % len(self.requiredMethodsDic)
            f.write('-' * 70 + '\n')
            f.write(fill(s) + '\n\n')

            mailtoLst = []
            for x in self.upToDateIdLst:
                obj = self.__translDic[x]
                if obj.note is None:
                    f.write('  ' + obj.classId + '\n')
                    mailtoLst.extend(self.__emails(obj.classId))

            fmail.write('up-to-date\n')
            fmail.write('; '.join(mailtoLst))


            # Write separately the list of "ALMOST up-to-date" translator classes.
            s = '''The following translator classes are ALMOST up-to-date (sorted
                alphabetically). This means that they derive from the
                Translator class, but there still may be some minor problems
                listed for them:'''
            f.write('\n' + ('-' * 70) + '\n')
            f.write(fill(s) + '\n\n')
            mailtoLst = []
            for x in self.upToDateIdLst:
                obj = self.__translDic[x]
                if obj.note is not None:
                    f.write('  ' + obj.classId + '\t-- ' + obj.note + '\n')
                    mailtoLst.extend(self.__emails(obj.classId))

            fmail.write('\n\nalmost up-to-date\n')
            fmail.write('; '.join(mailtoLst))

        # Write the list of the adapter based classes. The very obsolete
        # translators that derive from TranslatorEnglish are included.
        if self.adaptIdLst:
            s = '''The following translator classes need maintenance
                (the most obsolete at the end). The other info shows the
                estimation of Doxygen version when the class was last
                updated and number of methods that must be implemented to
                become up-to-date:'''
            f.write('\n' + '-' * 70 + '\n')
            f.write(fill(s) + '\n\n')

            # Find also whether some adapter classes may be removed.
            adaptMinVersion = '9.9.99'

            mailtoLst = []
            numRequired = len(self.requiredMethodsDic)
            for x in self.adaptIdLst:
                obj = self.__translDic[x]
                f.write('  %-30s' % obj.classId)
                f.write('  %-6s' % obj.readableStatus)
                numimpl = len(obj.missingMethods)
                pluralS = ''
                if numimpl > 1:
                    pluralS = 's'
                percent = 100 * numimpl / numRequired
                f.write('\t%2d method%s to implement (%d %%)' % (
                        numimpl, pluralS, percent))
                if obj.note:
                    f.write('\n\tNote: ' + obj.note + '\n')
                f.write('\n')
                mailtoLst.extend(self.__emails(obj.classId)) # to maintainer

                # Check the level of required adapter classes.
                if obj.status != '0.0.00' and obj.status < adaptMinVersion:
                    adaptMinVersion = obj.status

            fmail.write('\n\ntranslator based\n')
            fmail.write('; '.join(mailtoLst))

            # Set the note if some old translator adapters are not needed
            # any more.
            to_remove = {}
            for version, adaptClassId in list(self.adaptMethodsDic.values()):
                if version < adaptMinVersion:
                    to_remove[adaptClassId] = True

            if to_remove:
                lst = list(to_remove.keys())
                lst.sort()
                plural = len(lst) > 1
                note = 'Note: The adapter class'
                if plural:
                    note += 'es'
                note += ' ' + ', '.join(lst)
                if not plural:
                    note += ' is'
                else:
                    note += ' are'
                note += ' not used and can be removed.'
                f.write('\n' + fill(note) + '\n')

        # Write the list of the English-based classes.
        if self.EnBasedIdLst:
            s = '''The following translator classes derive directly from the
                TranslatorEnglish. The class identifier has the suffix 'En'
                that says that this is intentional. Usually, there is also
                a non-English based version of the translator for
                the language:'''
            f.write('\n' + '-' * 70 + '\n')
            f.write(fill(s) + '\n\n')

            for x in self.EnBasedIdLst:
                obj = self.__translDic[x]
                f.write('  ' + obj.classId)
                f.write('\timplements %d methods' % len(obj.implementedMethods))
                if obj.note:
                    f.write(' -- ' + obj.note)
                f.write('\n')

        # Check for not used translator methods and generate warning if found.
        # The check is rather time consuming.
        dic = self.__checkForNotUsedTrMethods()
        if dic:
            s = '''WARNING: The following translator methods are declared
                in the Translator class but their identifiers do not appear
                in source files. The situation should be checked. The .cpp
                files and .h files excluding the '*translator*' files
                in doxygen/src directory were simply searched for occurrence
                of the method identifiers:'''
            f.write('\n' + '=' * 70 + '\n')
            f.write(fill(s) + '\n\n')

            keys = list(dic.keys())
            keys.sort()
            for key in keys:
                f.write('  ' + dic[key] + '\n')
            f.write('\n')

        # Write the details for the translators.
        f.write('\n' + '=' * 70)
        f.write('\nDetails for translators (classes sorted alphabetically):\n')

        cls = list(self.__translDic.keys())
        cls.sort()

        for c in cls:
            obj = self.__translDic[c]
            assert(obj.classId != 'Translator')
            obj.report(f)

        if self.internal:
            f.write('\n\n@endverbatim\n')

        # Close the report file and the auxiliary file with e-mails.
        f.close()
        fmail.close()


    def __loadMaintainers(self):
        """Load and process the file with the maintainers.

        Fills the dictionary classId -> [(name, e-mail), ...]."""

        fname = os.path.join(self.org_doc_path, self.maintainersFileName)

        # Include the maintainers file to the group of files checked with
        # respect to the modification time.
        tim = os.path.getmtime(fname)
        if tim > self.lastModificationTime:
            self.lastModificationTime = tim

        # Process the content of the maintainers file.
        f = xopen(fname)
        inside = False  # inside the record for the language
        lineReady = True
        classId = None
        self.__maintainersDic = {}
        while lineReady:
            line = f.readline()            # next line
            lineReady = line != ''         # when eof, then line == ''

            line = line.strip()            # eof should also behave as separator
            if line != '' and line[0] == '%':    # skip the comment line
                continue

            if not inside:                # if outside of the record
                if line != '':            # should be language identifier
                    classId = line
                    inside = True
                # Otherwise skip empty line that do not act as separator.

            else:                         # if inside the record
                if line == '':            # separator found
                    inside = False
                else:
                    # If it is the first maintainer, create the empty list.
                    if classId not in self.__maintainersDic:
                        self.__maintainersDic[classId] = []

                    # Split the information about the maintainer and append
                    # the tuple. The address may be prefixed '[unreachable]'
                    # or whatever '[xxx]'. This will be processed later.
                    lst = line.split(':', 1)
                    assert(len(lst) == 2)
                    t = (lst[0].strip(), lst[1].strip())
                    self.__maintainersDic[classId].append(t)
        f.close()


    def generateLanguageDoc(self):
        """Checks the modtime of files and generates language.dox."""
        self.__loadMaintainers()

        # Check the last modification time of the VERSION file.
        fVerName = os.path.join(self.org_doxy_path, "VERSION")
        tim = os.path.getmtime(fVerName)
        if tim > self.lastModificationTime:
            self.lastModificationTime = tim

        # Check the last modification time of the template file. It is the
        # last file from the group that decide whether the documentation
        # should or should not be generated.
        fTplName = os.path.join(self.org_doc_path, self.languageTplFileName)
        tim = os.path.getmtime(fTplName)
        if tim > self.lastModificationTime:
            self.lastModificationTime = tim

        # If the generated documentation exists and is newer than any of
        # the source files from the group, do not generate it and quit
        # quietly.
        fDocName = os.path.join(self.doc_path, self.languageDocFileName)
        if os.path.isfile(fDocName):
            if os.path.getmtime(fDocName) > self.lastModificationTime:
                return

        # The document or does not exist or is older than some of the
        # sources. It must be generated again.
        #
        # Read the template of the documentation, and remove the first
        # attention lines.
        f = xopen(fTplName)
        doctpl = f.read()
        f.close()

        pos = doctpl.find('/***')
        assert pos != -1
        doctpl = doctpl[pos:]

        # Fill the tplDic by symbols that will be inserted into the
        # document template.
        tplDic = {}

        s = ('Do not edit this file. It was generated by the %s script.\n' +\
             ' * Edit the %s and %s files instead.') % (
             self.script_name, self.languageTplFileName, self.maintainersFileName)
        tplDic['editnote'] = s

        tplDic['doxVersion'] = self.doxVersion
        tplDic['supportedLangReadableStr'] = self.supportedLangReadableStr
        tplDic['translatorReportFileName'] = self.translatorReportFileName

        ahref = '<a href="' + self.translatorReportFileName
        ahref += '"\n><code>'  + self.translatorReportFileName + '</code></a>'
        tplDic['translatorReportLink'] = ahref
        tplDic['numLangStr'] = str(self.numLang)

        # Define templates for HTML table parts of the documentation.
        htmlTableTpl = '''
            \\latexonly
            \\footnotesize
            \\endlatexonly
            <table align="center" class=doxtable cellspacing="1" cellpadding="2">
            <tr>
              <th >Language</th>
              <th >Maintainer</th>
              <th >Contact address (replace the at and dot)</th>
              <th >Status</td>
              </tr>
              <!-- table content begin -->
            %s
              <!-- table content end -->
            </table>
            \\latexonly
            \\normalsize
            \\endlatexonly
            '''
        htmlTableTpl = textwrap.dedent(htmlTableTpl)
        htmlTrTpl = '\n  <tr bgcolor="#ffffff">%s\n  </tr>'
        htmlTdTpl = '\n    <td>%s</td>'
        htmlTdStatusColorTpl = '\n    <td bgcolor="%s">%s</td>'

        # Loop through transl objects in the order of sorted readable names
        # and add generate the content of the HTML table.
        trlst = []
        for name, obj in self.langLst:
            # Fill the table data elements for one row. The first element
            # contains the readable name of the language. Only the oldest
            # translators are color marked in the language column. Less
            # "heavy" color is used (when compared with the Status column).
            if obj.readableStatus.startswith('1.7'):
                bkcolor = self.getBgcolorByReadableStatus('1.7')
            elif obj.readableStatus.startswith('1.6'):
                bkcolor = self.getBgcolorByReadableStatus('1.6')
            elif obj.readableStatus.startswith('1.4'):
                bkcolor = self.getBgcolorByReadableStatus('1.4')
            else:
                bkcolor = '#ffffff'

            lst = [ htmlTdStatusColorTpl % (bkcolor, obj.langReadable) ]

            # The next two elements contain the list of maintainers
            # and the list of their mangled e-mails. For English-based
            # translators that are coupled with the non-English based,
            # insert the 'see' note.
            mm = None  # init -- maintainer
            ee = None  # init -- e-mail address
            if obj.status == 'En':
                # Check whether there is the coupled non-English.
                classId = obj.classId[:-2]
                if classId in self.__translDic:
                    lang = self.__translDic[classId].langReadable
                    mm = 'see the %s language' % lang
                    ee = '&#160;'

            if not mm and obj.classId in self.__maintainersDic:
                # Build a string of names separated by the HTML break element.
                # Special notes used instead of names are highlighted.
                lm = []
                for maintainer in self.__maintainersDic[obj.classId]:
                    name = maintainer[0]
                    if name.startswith('--'):
                        name = '<span style="color: red; background-color: yellow">'\
                               + name + '</span>'
                    lm.append(name)
                mm = '<br/>'.join(lm)

                # The marked addresses (they start with the mark '[unreachable]',
                # '[resigned]', whatever '[xxx]') will not be displayed at all.
                # Only the mark will be used instead.
                rexMark = re.compile('(?P<mark>\\[.*?\\])')
                le = []
                for maintainer in self.__maintainersDic[obj.classId]:
                    address = maintainer[1]
                    m = rexMark.search(address)
                    if m is not None:
                        address = '<span style="color: brown">'\
                                  + m.group('mark') + '</span>'
                    le.append(address)
                ee = '<br/>'.join(le)

            # Append the maintainer and e-mail elements.
            lst.append(htmlTdTpl % mm)
            lst.append(htmlTdTpl % ee)

            # The last element contains the readable form of the status.
            bgcolor = self.getBgcolorByReadableStatus(obj.readableStatus)
            lst.append(htmlTdStatusColorTpl % (bgcolor, obj.readableStatus.replace(".","\\.")))

            # Join the table data to one table row.
            trlst.append(htmlTrTpl % (''.join(lst)))

        # Join the table rows and insert into the template.
        htmlTable = htmlTableTpl % (''.join(trlst))

        # Put the HTML and LaTeX parts together and define the dic item.
        tplDic['informationTable'] = htmlTable

        # Insert the symbols into the document template and write it down.
        f = xopen(fDocName, 'w')
        f.write(doctpl % tplDic)
        f.close()

if __name__ == '__main__':

    # The Python 2.7+ or 3.3+ is required.
    major = sys.version_info[0]
    minor = sys.version_info[1]
    if (major == 2 and minor < 7) or (major == 3 and minor < 0):
        print('Python 2.7+ or Python 3.0+ are required for the script')
        sys.exit(1)

    # The translator manager builds the Transl objects, parses the related
    # sources, and keeps them in memory.
    trMan = TrManager()

    # Process the Transl objects and generate the output files.
    trMan.generateLanguageDoc()
    trMan.generateTranslatorReport()



import sys
do_print = False
for line in sys.stdin:
    if line.startswith("%--- Begin generated contents ---"):
        do_print = True
    elif line.startswith("%--- End generated contents ---"):
        do_print = False
    elif do_print:
        sys.stdout.write(line)
